# -*- coding: utf-8 -*-
"""
multi_analysis class

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
from joblib import Parallel, delayed
from copy import deepcopy

from .dfc_methods import *
from .dfc_utils import get_subj_ts_dict, common_subj_lst

############################# MultiAnalysis class ################################

"""

todo:
- 
"""

class MultiAnalysis:
    # if self.n_jobs is None => no parallelization

    def __init__(self, analysis_name='', **params):

        self.analysis_name = analysis_name
        
        self.params = params
        if not 'n_jobs' in self.params:
            self.params['n_jobs'] = -1
        if not 'verbose' in self.params:
            self.params['verbose'] = 1
        if not 'backend' in self.params:
            self.params['backend'] = 'loky'

        self.MEASURES_lst_ = None
        self.MEASURES_fit_lst_ = []
        self.MEASURES_name_lst = []
        self.params_methods = {}
        self.alter_hparams = {}
        self.hyper_param_info = {}

    @property
    def MEASURES_lst(self):
        assert not self.MEASURES_lst_ is None, \
            'first set the MEASURES_lst!'
        return self.MEASURES_lst_

    @property
    def MEASURES_fit_lst(self):
        return self.MEASURES_fit_lst_

    def set_MEASURES_lst(self, MEASURES_lst):
        self.MEASURES_lst_ = MEASURES_lst

    def set_MEASURES_fit_lst(self, MEASURES_fit_lst):
        self.MEASURES_fit_lst_ = MEASURES_fit_lst

    def measures_initializer(self, MEASURES_name_lst, params_methods, alter_hparams):

        '''
        - this will test values in hyper_params other than
            values already in self.params. values in self.params 
            will be considered the reference
        sample:
        hyper_params = { \
            'n_states': [6, 12, 16], \
            'normalization': [True], \
            'num_subj': [50, 100, 395], \
            'num_select_nodes': [50, 100, 333], \
            'num_time_point': [500, 800, 1200], \
            'Fs_ratio': [0.50, 1.00, 1.50], \
            'noise_ratio': [0.00, 0.50, 1.00], \
            'num_realization': [1, 2, 3], \
            }
            
            MEASURES_name_lst = ( \
                'SlidingWindow', \
                'Time-Freq', \
                'CAP', \
                'ContinuousHMM', \
                'Windowless', \
                'Clustering', \
                'DiscreteHMM' \
                )
        '''

        self.MEASURES_name_lst = MEASURES_name_lst
        self.params_methods = params_methods
        self.alter_hparams = alter_hparams

        # a list of MEASURES with default parameter values
        MEASURES_lst = self.create_measure_obj(MEASURES_name_lst=MEASURES_name_lst, **params_methods)

        # adding MEASURES with alternative parameter values
        hyper_param_info = {}
        hyper_param_info['default_values'] = params_methods
        for hyper_param in alter_hparams:
            for value in alter_hparams[hyper_param]:
                params = deepcopy(params_methods)
                params[hyper_param] = value
                hyper_param_info[hyper_param+'_'+str(value)] = deepcopy(params)
                new_MEASURES = self.create_measure_obj(MEASURES_name_lst=MEASURES_name_lst, **params)
                for new_measure in new_MEASURES:
                    flag=0
                    for MEASURE in MEASURES_lst:
                        if new_measure.issame(MEASURE):
                            flag=1
                    if flag==0:
                        MEASURES_lst.append(new_measure)

        self.hyper_param_info = hyper_param_info

        return MEASURES_lst

    def create_measure_obj(self, MEASURES_name_lst, **params):

        MEASURES_lst = list()
        for MEASURES_name in MEASURES_name_lst:

            ###### CAP ######
            if MEASURES_name=='CAP':
                measure = CAP(**params)

            ###### CONTINUOUS HMM ######
            if MEASURES_name=='ContinuousHMM':
                measure = HMM_CONT(**params)

            ###### WINDOW_LESS ######
            if MEASURES_name=='Windowless':
                measure = WINDOWLESS(**params)

            ###### SLIDING WINDOW ######
            if MEASURES_name=='SlidingWindow':
                measure = SLIDING_WINDOW(**params)

            ###### TIME FREQUENCY ######
            if MEASURES_name=='Time-Freq':
                measure = TIME_FREQ(**params)

            ###### SLIDING WINDOW + CLUSTERING ######
            if MEASURES_name=='Clustering':
                measure = SLIDING_WINDOW_CLUSTR(**params)

            ###### DISCRETE HMM ######
            if MEASURES_name=='DiscreteHMM':
                measure = HMM_DISC(**params)

            MEASURES_lst.append(measure)

        return MEASURES_lst

    def SB_MEASURES_lst(self, MEASURES_lst): # returns state_based measures
        SB_MEASURES = list()
        for measure in MEASURES_lst:
            if measure.is_state_based:
                SB_MEASURES.append(measure)
        return SB_MEASURES

    def DD_MEASURES_lst(self, MEASURES_lst): # returns data_driven measures
        DD_MEASURES = list()
        for measure in MEASURES_lst:
            if not measure.is_state_based:
                DD_MEASURES.append(measure)
        return DD_MEASURES

    ##################### FCS ESTIMATION ######################

    def estimate_group_FCS(self, time_series_dict):

        # time_series_dict is a dict of time_series

        for session in time_series_dict:

            time_series = time_series_dict[session]
            SB_MEASURES_lst = self.SB_MEASURES_lst(self.MEASURES_lst)
            if self.params['n_jobs'] is None:
                SB_MEASURES_lst_NEW = list()
                for measure in SB_MEASURES_lst:
                    SB_MEASURES_lst_NEW.append( \
                        measure.estimate_FCS(time_series=time_series) \
                        )
            else:
                SB_MEASURES_lst_NEW = Parallel( \
                    n_jobs=self.params['n_jobs'], verbose=self.params['verbose'], backend=self.params['backend'])( \
                    delayed(measure.estimate_FCS)(time_series=time_series) \
                        for measure in SB_MEASURES_lst)
            self.MEASURES_fit_lst_[session] = self.DD_MEASURES_lst(self.MEASURES_lst) + SB_MEASURES_lst_NEW

    ##################### dFC ASSESSMENT ######################

    def group_dFC_assess(self, time_series_dict):

        # time_series_dict is a dict of time_series

        SUBJ_s_dFC_dict = {}
        
        SUBJECTs = common_subj_lst(time_series_dict) 

        if self.params['n_jobs'] is None:
            OUT = list()
            for subject in SUBJECTs:
                OUT.append( \
                    self.subj_lvl_dFC_assess( \
                    time_series_dict=get_subj_ts_dict(time_series_dict, subjs_id=subject), \
                    ))
        else:
            OUT = Parallel( \
                        n_jobs=self.params['n_jobs'], \
                        verbose=self.params['verbose'], \
                        backend=self.params['backend'])( \
                    delayed(self.subj_lvl_dFC_assess)( \
                        time_series_dict=get_subj_ts_dict(time_series_dict, subjs_id=subject), \
                        ) \
                        for subject in SUBJECTs)
        
        return OUT

    def subj_lvl_dFC_assess(self, time_series_dict):

        # time_series_dict is a dict of time_series

        dFC_dict = {}
        # dFC_corr_assess_dict = {}

        if self.params['n_jobs'] is None:
            dFC_lst = list()
            for measure in self.MEASURES_fit_lst_:
                dFC_lst.append( \
                    measure.estimate_dFC(time_series=time_series_dict[measure.params['session']]) \
                )
        else:
            dFC_lst = Parallel( \
                n_jobs=self.params['n_jobs'], verbose=self.params['verbose'], backend=self.params['backend'])( \
                delayed(measure.estimate_dFC)(time_series=time_series_dict[measure.params['session']]) \
                    for measure in self.MEASURES_fit_lst_)

        dFC_dict['dFC_lst'] = dFC_lst

        return dFC_dict

##############################################################################################################


"""
Implementation of dFC methods.

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
import time
from sklearn.cluster import KMeans
from pyclustering.cluster.kmeans import kmeans
from pyclustering.cluster.center_initializer import kmeans_plusplus_initializer
from pyclustering.utils.metric import distance_metric, type_metric

from .base_dfc_method import BaseDFCMethod
from ..time_series import TIME_SERIES
from ..dfc import DFC
from .sliding_window import SLIDING_WINDOW
from .time_freq import TIME_FREQ
from ..dfc_utils import dFC_mat2vec, dFC_vec2mat

########################### Sliding_Window + Clustering ############################

"""
- We used a tapered window as in Allen et al., created by convolving a rectangle (width = 22 TRs = 44s) 
  with a Gaussian (σ = 3 TRs) and slid in steps of 1 TR, resulting in W= 126 windows (Allen et al., 2014).
- Kmeans Clustering is repeated 500 times to escape local minima (Allen et al., 2014)
- for clustering, we have a 2-level kmeans clustering. First, we cluster FCSs of each subject. Then, we
    cluster all clustering centers from all subjects. the final estimate_dFC is using the second kmeans
    model (Allen et al., 2014; Ou et al., 2015). 

Parameters
    ----------
    y1, y2 : numpy.ndarray, list
        Input signals.
    dt : float
        Sample spacing.

todo:
- pyclustering(manhattan) has a problem when suing predict
"""

class SLIDING_WINDOW_CLUSTR(BaseDFCMethod):

    def __init__(self, clstr_distance='euclidean', **params):

        assert clstr_distance=='euclidean' or clstr_distance=='manhattan', \
            "Clustering distance not recognized. It must be either \
                euclidean or manhattan."
        self.logs_ = ''
        self.TPM = []
        self.FCS_ = []
        self.mean_act = []
        self.FCS_fit_time_ = None
        self.dFC_assess_time_ = None

        self.params_name_lst = ['measure_name', 'is_state_based', 'clstr_base_measure', 'sw_method', 'tapered_window',
            'clstr_distance', 'coi_correction',
            'n_subj_clstrs', 'W', 'n_overlap', 'n_states', 'normalization',
            'n_jobs', 'verbose', 'backend',
            'num_subj', 'num_select_nodes', 'num_time_point', 'Fs_ratio',
            'noise_ratio', 'num_realization', 'session']
        self.params = {}
        for params_name in self.params_name_lst:
            if params_name in params:
                self.params[params_name] = params[params_name]
            else:
                self.params[params_name] = None
        
        self.params['measure_name'] = 'Clustering'
        self.params['is_state_based'] = True
        self.params['clstr_distance'] = clstr_distance

        assert self.params['clstr_base_measure'] in self.base_methods_name_lst, \
            "Base method not recognized."

    @property
    def measure_name(self):
        return self.params['measure_name'] #+ '_' + self.base_method

    def dFC_mat2vec(self, C_t):
        return dFC_mat2vec(C_t)
        # if len(C_t.shape)==2:
        #     assert C_t.shape[0]==C_t.shape[1],\
        #         'C is not a square matrix'
        #     return C_t[np.triu_indices(C_t.shape[1], k=0)]

        # F = list()
        # for t in range(C_t.shape[0]):
        #     C = C_t[t, : , :]
        #     assert C.shape[0]==C.shape[1],\
        #         'C is not a square matrix'
        #     F.append(C[np.triu_indices(C_t.shape[1], k=0)])

        # F = np.array(F)
        # return F

    def dFC_vec2mat(self, F, N):
        return dFC_vec2mat(F=F, N=N)
        # C = list()
        # iu = np.triu_indices(N, k=0)
        # for i in range(F.shape[0]):
        #     K = np.zeros((N, N))
        #     K[iu] = F[i,:]
        #     K = K + np.multiply(K.T, 1-np.eye(N))
        #     C.append(K)
        # C = np.array(C)
        # return C

    def clusters_lst2idx(self, clusters):
        Z = np.zeros((self.F.shape[0],))
        for i, cluster in enumerate(clusters):
            for sample in cluster:
                Z[sample] = i
        return Z.astype(int)

    def cluster_FC(self, FCS_raw, n_clusters, n_regions):

        F = self.dFC_mat2vec(FCS_raw)

        if self.params['clstr_distance']=='manhattan':
            pass
            # ########### Manhattan Clustering ##############
            # # Prepare initial centers using K-Means++ method.
            # initial_centers = kmeans_plusplus_initializer(F, self.n_states).initialize()
            # # create metric that will be used for clustering
            # manhattan_metric = distance_metric(type_metric.MANHATTAN)
            # # Create instance of K-Means algorithm with prepared centers.
            # kmeans_ = kmeans(F, initial_centers, metric=manhattan_metric)
            # # Run cluster analysis and obtain results.
            # kmeans_.process()
            # Z = self.clusters_lst2idx(kmeans_.get_clusters())
            # F_cent = np.array(kmeans_.get_centers())
        else:
            ########### Euclidean Clustering ##############
            kmeans_ = KMeans(n_clusters=n_clusters, n_init=500).fit(F)
            F_cent = kmeans_.cluster_centers_

        FCS_ = self.dFC_vec2mat(F_cent, N=n_regions)
        return FCS_, kmeans_

        
    def estimate_FCS(self, time_series):

        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        time_series = self.manipulate_time_series4FCS(time_series)

        # start timing
        tic = time.time()

        base_dFC = None
        if self.params['clstr_base_measure']=='Time-Freq':
            base_dFC = TIME_FREQ(**self.params)
        if self.params['clstr_base_measure']=='SlidingWindow':
            base_dFC = SLIDING_WINDOW(**self.params)

        # 1-level clustering
        # dFC_raw = base_dFC.estimate_dFC( \
        #         time_series=time_series \
        #         )
        # self.FCS_, self.kmeans_ = self.cluster_FC( \
        # dFC_raw.get_dFC_mat(TRs=self.dFC_raw.TR_array), \
        # n_regions = dFC_raw.n_regions \
        # )

        # 2-level clustering
        SUBJECTs = time_series.subj_id_lst
        FCS_1st_level = None
        SW_dFC = None
        for subject in SUBJECTs:
            
            dFC_raw = base_dFC.estimate_dFC( \
                time_series=time_series.get_subj_ts(subjs_id=subject) \
                )

            # test
            if dFC_raw.n_time<self.params['n_subj_clstrs']:
                print( \
                    'Number of subject-level clusters cannot be more than SW dFC samples! n_subj_clstrs was changed to ' \
                        + str(dFC_raw.n_time) + '. This change will cause problems in similarity assessment.')
                self.params['n_subj_clstrs'] = dFC_raw.n_time

            FCS, _ = self.cluster_FC( \
                FCS_raw = dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array), \
                n_clusters = self.params['n_subj_clstrs'], \
                n_regions = dFC_raw.n_regions \
                )
            
            if SW_dFC is None:
                SW_dFC = dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array)
            else:
                SW_dFC = np.concatenate((SW_dFC, dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array)), axis=0)
            if FCS_1st_level is None:
                FCS_1st_level = FCS
            else:
                FCS_1st_level = np.concatenate((FCS_1st_level, FCS), axis=0)
        
        self.FCS_, self.kmeans_ = self.cluster_FC( \
            FCS_raw=FCS_1st_level, \
            n_clusters = self.params['n_states'], \
            n_regions = dFC_raw.n_regions \
            )
        self.Z = self.kmeans_.predict(self.dFC_mat2vec(SW_dFC))

        # mean activation of states
        self.set_mean_activity(time_series)

        # record time
        self.set_FCS_fit_time(time.time() - tic)

        return self

    def estimate_dFC(self, time_series):
        
        assert type(time_series) is TIME_SERIES, \
            "time_series must be of TIME_SERIES class."

        assert len(time_series.subj_id_lst)==1, \
            'this function takes only one subject as input.'

        time_series = self.manipulate_time_series4dFC(time_series)

        # start timing
        tic = time.time()

        base_dFC = None
        if self.params['clstr_base_measure']=='Time-Freq':
            base_dFC = TIME_FREQ(**self.params)
        if self.params['clstr_base_measure']=='SlidingWindow':
            base_dFC = SLIDING_WINDOW(**self.params)
                    
        dFC_raw = base_dFC.estimate_dFC(time_series=time_series)

        F = self.dFC_mat2vec(dFC_raw.get_dFC_mat(TRs=dFC_raw.TR_array))

        if self.params['clstr_distance']=='manhattan':
            pass
            # ########### Manhattan Clustering ##############
            # self.kmeans_.predict(F)
            # Z = self.clusters_lst2idx(self.kmeans_.get_clusters())
        else:
            ########### Euclidean Clustering ##############
            Z = self.kmeans_.predict(F)

        # record time
        self.set_dFC_assess_time(time.time() - tic)

        dFC = DFC(measure=self)
        dFC.set_dFC(FCSs=self.FCS_, \
            FCS_idx=Z, \
            TS_info=time_series.info_dict, \
            TR_array=dFC_raw.TR_array \
            )

        return dFC


################################################################################


from functions.dFC_funcs import *
import numpy as np
import time
import hdf5storage
import scipy.io as sio
import os
os.environ["MKL_NUM_THREADS"] = '64'
os.environ["NUMEXPR_NUM_THREADS"] = '64'
os.environ["OMP_NUM_THREADS"] = '64'

print('################################# subject-level dFC assessment CODE started running ... #################################')

################################# Parameters #################################
###### DATA PARAMETERS ######

input_root = './'
output_root = './'

################################# LOAD #################################

dFC_analyzer = np.load(input_root+'dFC_analyzer.npy',allow_pickle='TRUE').item()
data_loader = np.load(input_root+'data_loader.npy',allow_pickle='TRUE').item()

################################# LOAD FIT MEASURES #################################

if dFC_analyzer.MEASURES_fit_lst==[]:
    ALL_RECORDS = os.listdir(input_root+'fitted_MEASURES/')
    ALL_RECORDS = [i for i in ALL_RECORDS if 'MEASURE' in i]
    ALL_RECORDS.sort()
    MEASURES_fit_lst = list()
    for s in ALL_RECORDS:
        fit_measure = np.load(input_root+'fitted_MEASURES/'+s, allow_pickle='TRUE').item()
        MEASURES_fit_lst.append(fit_measure)
    dFC_analyzer.set_MEASURES_fit_lst(MEASURES_fit_lst)
    print('fitted MEASURES loaded ...')
    # np.save('./dFC_analyzer.npy', dFC_analyzer) 

################################# LOAD DATA #################################

task_id = int(os.getenv("SGE_TASK_ID"))
subj_id = data_loader.SUBJECTS[task_id-1] # SGE_TASK_ID starts from 1 not 0

BOLD = data_loader.load(subj_id2load=subj_id)

################################# dFC ASSESSMENT #################################

tic = time.time()
print('Measurement Started ...')

print("dFCM estimation started...")
dFCM_dict = dFC_analyzer.subj_lvl_dFC_assess(time_series_dict=BOLD)
print("dFCM estimation done.")

print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

################################# SAVE DATA #################################

# folder = output_root+'dFC_assessed/SUBJ_'+str(subj_id)
# if not os.path.exists(folder):
#     os.makedirs(folder)

# for dFCM_id, dFCM in enumerate(dFCM_dict['dFCM_lst']):
#     np.save(folder+'/dFCM_'+str(dFCM_id)+'.npy', dFCM) 

################################# SIMILARITY MEASUREMENT #################################

similarity_assessment = SIMILARITY_ASSESSMENT(dFCM_lst=dFCM_dict['dFCM_lst'])

tic = time.time()
print('Measurement Started ...')

print("Similarity measurement started...")
SUBJ_output = similarity_assessment.run(FILTERS=dFC_analyzer.hyper_param_info, downsampling_method='default')
print("Similarity measurement done.")

print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

# Save
folder = output_root+'similarity_measured'
if not os.path.exists(folder):
    os.makedirs(folder)

np.save(folder+'/SUBJ_'+str(subj_id)+'_output.npy', SUBJ_output) 

#################################################################################


from calendar import c
from pydfc import (
    data_loader,
    MultiAnalysis,
)
import time
import numpy as np
import os
import warnings

warnings.simplefilter('ignore')

os.environ["MKL_NUM_THREADS"] = '16'
os.environ["NUMEXPR_NUM_THREADS"] = '16'
os.environ["OMP_NUM_THREADS"] = '16'

################################# Parameters #################################

# Data parameters
# main_root = '../../DATA/ds002785/' # for local
main_root = '../../../DATA/task-based/openneuro/ds002785/' # for server

# subjects used for dFC assessment do not need to be the same as those used for FCS_estimate
# you can set the new roi root and data load parameters here:
roi_root = f"{main_root}/derivatives/ROI_timeseries"
fitted_measures_root = f"{main_root}/derivatives/fitted_MEASURES"
output_root = f"{main_root}/derivatives/dFC_assessed"

# for consistency we use 0 for resting state
TASKS = [
    'task-restingstate', 
    'task-anticipation', 
    'task-emomatching', 
    'task-faces', 
    'task-gstroop', 
    'task-workingmemory'
]

# find all subjects across all tasks
SUBJECTS = data_loader.find_subj_list(data_root=roi_root, sessions=TASKS)

# job_id selects the subject
job_id = int(os.getenv("SGE_TASK_ID"))
if job_id > len(SUBJECTS):
    print('job_id > len(SUBJECTS)')
    exit()
subj_id = SUBJECTS[job_id-1] # SGE_TASK_ID starts from 1 not 0

for task in TASKS:

    MA = np.load(f"{fitted_measures_root}/{task}/multi_analysis.npy", allow_pickle='TRUE').item()

    # check if the subject has this task
    SUBJECTS_with_this_task = data_loader.find_subj_list(data_root=roi_root, sessions=[task])
    if not subj_id in SUBJECTS_with_this_task:
        print(f"subject {subj_id} not in the list of subjects with task {task}")
        continue

    ################################# LOAD FIT MEASURES #################################

    ALL_RECORDS = os.listdir(f"{fitted_measures_root}/{task}/")
    ALL_RECORDS = [i for i in ALL_RECORDS if 'MEASURE' in i]
    ALL_RECORDS.sort()
    MEASURES_fit_lst = list()
    for s in ALL_RECORDS:
        fit_measure = np.load(f"{fitted_measures_root}/{task}/{s}", allow_pickle='TRUE').item()
        MEASURES_fit_lst.append(fit_measure)
    MA.set_MEASURES_fit_lst(MEASURES_fit_lst)
    print('fitted MEASURES loaded ...')

    ################################# LOAD DATA #################################
    
    print(f"subject-level dFC assessment CODE started running ... for task {task} of subject {subj_id} ...")

    BOLD = data_loader.load_TS(
            data_root=roi_root, 
            file_name='time_series.npy', 
            SESSIONs=[task], 
            subj_id2load=subj_id
    )

    ################################# dFC ASSESSMENT #################################

    tic = time.time()
    print('Measurement Started ...')

    print("dFC estimation started...")
    dFC_dict = MA.subj_lvl_dFC_assess(time_series_dict=BOLD)
    print("dFC estimation done.")

    print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

    ################################# SAVE DATA #################################

    folder = f"{output_root}/{task}/{subj_id}"
    if not os.path.exists(folder):
        os.makedirs(folder)

    for dFC_id, dFC in enumerate(dFC_dict['dFC_lst']):
        np.save(f"{folder}/dFC_{str(dFC_id)}.npy", dFC) 
    
#######################################################################################


from functions.dFC_funcs import *
import numpy as np
import time
import hdf5storage
import scipy.io as sio
import os
os.environ["MKL_NUM_THREADS"] = '64'
os.environ["NUMEXPR_NUM_THREADS"] = '64'
os.environ["OMP_NUM_THREADS"] = '64'

################################# Parameters #################################

###### DATA PARAMETERS ######
DATA_type = 'Gordon' # 'Gordon' or 'simulated' or 'ICA'

output_root = './../../../../../RESULTs/methods_implementation/'
# output_root = '/data/origami/dFC/RESULTs/methods_implementation/'
# output_root = '/Users/mte/Documents/McGill/Project/dFC/RESULTs/methods_implementation/'

data_root_simul = './../../../../DATA/TVB data/'
data_root_sample = './sampleDATA/'
data_root_gordon = './../../../../DATA/HCP/HCP_Gordon/'
data_root_ica = './../../../../DATA/HCP/HCP_PTN1200/node_timeseries/3T_HCP1200_MSMAll_d50_ts2/'

num_subj = 2
select_nodes = True
rand_node_slct = False
num_select_nodes = 50

###### MEASUREMENT PARAMETERS ######

# W is in sec

params_methods = { \
    # Sliding Parameters
    'W': 44, 'n_overlap': 0.5, \
    # State Parameters
    'n_states': 6, 'n_subj_clstrs': 20, 'n_hid_states': 4, \
    # Parallelization Parameters
    'n_jobs': 2, 'verbose': 0, 'backend': 'loky' \
}

###### SIMILARITY PARAMETERS ######

sim_assess_params= { \
    'run_analysis': True, \
    'num_samples': 100, \
    'matching_method': 'score', \
    'n_jobs': 8, 'backend': 'loky' \
}

###### SIMILARITY PARAMETERS ######

dyn_conn_det_params = { \
    'run_analysis': False, \
    'N': 30, 'L': 1200, 'p': 100, \
    'n_jobs': 8, 'backend': 'loky' \
}

###### dFC ANALYZER PARAMETERS ######

params_dFC_analyzer = { \
    # VISUALIZATION
    'vis_TR_idx': list(range(10, 20, 1)),'save_image': True, 'output_root': output_root, \
    # Parallelization Parameters
    'n_jobs': 8, 'verbose': 0, 'backend': 'loky', \
    # Similarity Assessment Parameters
    'sim_assess_params': sim_assess_params, \
    # Dynamic Connection Detector Parameters
    'dyn_conn_det_params': dyn_conn_det_params \
}


################################# FINDING SUBJECTs LIST #################################

# ICA
ALL_RECORDS = os.listdir(data_root_ica)
ALL_RECORDS = [i for i in ALL_RECORDS if '.txt' in i]
ALL_RECORDS.sort()
SUBJECTS_ica = list()
for s in ALL_RECORDS:
    num = s[:s.find('.')]
    SUBJECTS_ica.append(num)
SUBJECTS_ica = list(set(SUBJECTS_ica))
SUBJECTS_ica.sort()

# GORDON
ALL_RECORDS = os.listdir(data_root_gordon)
ALL_RECORDS = [i for i in ALL_RECORDS if 'Rest' in i]
ALL_RECORDS.sort()
SUBJECTS_gordon = list()
for s in ALL_RECORDS:
    num = s[:s.find('_')]
    SUBJECTS_gordon.append(num)
SUBJECTS_gordon = list(set(SUBJECTS_gordon))
SUBJECTS_gordon.sort()

SUBJECTS = intersection(SUBJECTS_gordon, SUBJECTS_ica)

print( str(len(SUBJECTS)) + ' subjects were found. ' + str(num_subj) + ' subjects were selected.')

SUBJECTS = SUBJECTS[0:num_subj]

################################# Load ICA BOLD data (HCP) #################################

if DATA_type=='ICA':

    SESSIONs = ['session_1']
    BOLD_Fs = 1/0.72

    BOLD = {}
    for session in SESSIONs:
        BOLD[session] = None
        for subject in SUBJECTS:
            time_series = np.loadtxt( \
                data_root_ica + subject + '.txt', dtype='float64' \
                )
            time_series = time_series.T
            
            # time_series = time_series - np.repeat(np.mean(time_series, axis=1)[:,None], time_series.shape[1], axis=1) # ???????????????????????

            if BOLD[session] is None:
                BOLD[session] = TIME_SERIES(data=time_series, subj_id=subject, Fs=BOLD_Fs, TS_name='BOLD ICA', session_name=session)
            else:
                BOLD[session].append_ts(new_time_series=time_series, subj_id=subject)

        print(BOLD[session].n_regions, BOLD[session].n_time)

################################# Load Real BOLD data (HCP) #################################

if DATA_type=='Gordon':

    SESSIONs = ['Rest1_LR' , 'Rest1_RL', 'Rest2_LR', 'Rest2_RL']
    BOLD_Fs = 1/0.72

    # LOAD Region Location DATA

    locs = sio.loadmat(data_root_gordon+'Gordon333_LOCS.mat')
    locs = locs['locs']

    # LOAD Region Data

    file = data_root_gordon+'Gordon333_Key.txt'
    f = open(file, 'r')

    atlas_data = []
    for line in f:
        row = line.split()
        atlas_data.append(row)

    BOLD = {}
    for session in SESSIONs:
        BOLD[session] = None
        for subject in SUBJECTS:

            subj_fldr = subject + '_' + session

            # LOAD BOLD Data

            DATA = hdf5storage.loadmat(data_root_gordon+subj_fldr+'/ROI_data_Gordon_333_surf.mat')
            time_series = DATA['ROI_data']

            time_series = time_series.T

            # time_series = time_series - np.repeat(np.mean(time_series, axis=1)[:,None], time_series.shape[1], axis=1) # ???????????????????????

            for n in range(time_series.shape[0]):
                time_series[n, :] = time_series[n, :] - np.mean(time_series[n, :])
                time_series[n, :] = np.divide(time_series[n, :], np.std(time_series[n, :]))

            if BOLD[session] is None:
                BOLD[session] = TIME_SERIES(data=time_series, subj_id=subject, \
                                    Fs=BOLD_Fs, \
                                    locs=locs, nodes_info=atlas_data, \
                                    TS_name='BOLD Real', session_name=session \
                                )
            else:
                BOLD[session].append_ts(new_time_series=time_series, subj_id=subject)

        print( '*** Session ' + session + ': ' )
        print( 'number of regions= '+str(BOLD[session].n_regions) + ', number of TRs= ' + str(BOLD[session].n_time) )


        # select nodes

        if select_nodes:
            if rand_node_slct:
                nodes_idx = np.random.choice(range(BOLD[session].n_regions), size=num_select_nodes, replace=False)
                nodes_idx.sort()
            else:
                nodes_idx = np.array(list(range(47, 88)) + list(range(224, 263)))
            BOLD[session].select_nodes(nodes_idx=nodes_idx)

            print( 'number of regions selected= ' + str(BOLD[session].n_regions) )

################################# Load Sample BOLD data #################################

if DATA_type=='sample':

    BOLD_Fs = 1/0.5

    ###### BOLD DATA ######
    time_BOLD = np.load(data_root_sample+'bold_time.npy')/1e3    
    time_series = np.load(data_root_sample+'bold_data.npy')

    time_series = time_series.T

    BOLD = None
    for subject in range(5):

        if BOLD is None:
            BOLD = TIME_SERIES( \
                data=time_series[:, (subject)*1200:(subject+1)*1200], \
                subj_id=str(subject+1), Fs=BOLD_Fs, \
                time_array=time_BOLD[(subject)*1200:(subject+1)*1200], \
                TS_name='BOLD Sample' \
                )
        else:
            BOLD.append_ts( \
                new_time_series=time_series[:, (subject)*1200:(subject+1)*1200], \
                time_array=time_BOLD[(subject)*1200:(subject+1)*1200],
                subj_id=str(subject+1) \
                )

    print(BOLD.n_regions, BOLD.n_time)
        

################################# Load Simulated BOLD data #################################

if DATA_type=='simulated':
    time_BOLD = np.load(data_root_simul+'bold_time.npy')/1e3    
    time_series_BOLD = np.load(data_root_simul+'bold_data.npy')

    BOLD = TIME_SERIES(data=time_series_BOLD.T, subj_id='1', Fs=BOLD_Fs, time_array=time_BOLD, TS_name='BOLD Simulation')

################################# Load Simulated Tavg data #################################

if DATA_type=='simulated':
    time_Tavg = np.load(data_root_simul+'TVB data/tavg_time.npy')/1e3    
    time_series_Tavg = np.load(data_root_simul+'TVB data/tavg_data.npy')

    TAVG = TIME_SERIES(data=time_series_Tavg.T, subj_id='1', Fs=200, time_array=time_Tavg, TS_name='Tavg Simulation')

################################# Visualize BOLD #################################

for session in BOLD:
    BOLD[session].visualize(start_time=0, end_time=50, nodes_lst=list(range(10)), \
        save_image=params_dFC_analyzer['save_image'], output_root=output_root+'BOLD_signal_'+session)

################################# Measure dFC #################################

###### CONTINUOUS HMM ######
hmm_cont = HMM_CONT(**params_methods)

###### WINDOW_LESS ######
windowless = WINDOWLESS(**params_methods)

###### SLIDING WINDOW ######
sw_pc = SLIDING_WINDOW(sw_method='pear_corr', **params_methods)
sw_mi = SLIDING_WINDOW(sw_method='MI', **params_methods)

###### TIME FREQUENCY ######
time_freq_cwt = TIME_FREQ(method='CWT_mag', **params_methods)
time_freq_wtc = TIME_FREQ(method='WTC', **params_methods)

###### SLIDING WINDOW + CLUSTERING ######
swc_pc = SLIDING_WINDOW_CLUSTR(base_method='pear_corr', **params_methods)

###### DISCRETE HMM ######
hmm_disc_pc = HMM_DISC(base_method='pear_corr', **params_methods)


MEASURES = [

    hmm_cont, \

    windowless, \

    sw_pc, \
    # sw_mi, \
    # sw_gLasso, \

    # time_freq_cwt, \
    # time_freq_cwt_r, \
    # time_freq_wtc, \

    swc_pc, \
    # swc_gLasso, \

    hmm_disc_pc,\
    # hmm_disc_gLasso, \

]

dFC_analyzer = DFC_ANALYZER(MEASURES_lst=MEASURES, \
    analysis_name='reproducibility assessment', \
    **params_dFC_analyzer \
)

tic = time.time()
print('Measurement Started ...')
SUBJs_dFC_session_sim_dict = dFC_analyzer.analyze(time_series_dict=BOLD)

# Save
np.save('./dFC_analyzer.npy', dFC_analyzer) 
np.save('./dFC_session_sim.npy', SUBJs_dFC_session_sim_dict) 

print('Measurement required %0.3f seconds.' % (time.time() - tic, ))

################################# SIMILARITY ASSESSMENT #################################

# print_dict(dFC_analyzer.methods_corr)
dFC_analyzer.similarity_analyze(SUBJs_dFC_session_sim_dict)

################################# STATE MATCH #################################

state_match = dFC_analyzer.state_match()
# Save
np.save('./state_match.npy', state_match) 

################################# POST ANALYSIS #################################

dFC_analyzer.post_analyze()

################################# SAMPLE CHECK #################################

session = 'Rest1_LR'

measure_pairs = [['ContinuousHMM', 'Windowless'], ['ContinuousHMM', 'Clustering_pear_corr'], ['ContinuousHMM', 'DiscreteHMM_pear_corr']]
subj_id = np.unique(BOLD[session].subj_id_array)[1]
time_series = get_subj_ts_dict(BOLD, subj_id=subj_id)[session]

# if params_dFC_analyzer['n_jobs'] is None:
#     dFCM_lst = list()
#     for measure in dFC_analyzer.MEASURES_fit_lst[session]:
#         dFCM_lst.append( \
#             measure.estimate_dFCM(time_series=time_series) \
#         )
# else:
#     dFCM_lst = Parallel( \
#         n_jobs=params_dFC_analyzer['n_jobs'], verbose=params_dFC_analyzer['verbose'], backend=params_dFC_analyzer['backend'])( \
#         delayed(measure.estimate_dFCM)(time_series=time_series) \
#             for measure in dFC_analyzer.MEASURES_fit_lst[session])

for measure_pair in measure_pairs:
    measure_1 = dFC_analyzer.MEASURES_fit_dict[session][measure_pair[0]]
    measure_2 = dFC_analyzer.MEASURES_fit_dict[session][measure_pair[1]]

    # re-run estimate_dFCM for a sample subject

    print("dFCM estimation started...")
    dFCM_i = measure_1.estimate_dFCM(time_series=time_series)
    print("dFCM estimation done.")

    print("dFCM estimation started...")
    dFCM_j = measure_2.estimate_dFCM(time_series=time_series)
    print("dFCM estimation done.")

    # TRs = TR_intersection(dFCM_lst)
    TRs = TR_intersection([dFCM_i, dFCM_j])

    # state match visualization on the sample subject with the state matching result from all subjects
    dFC_analyzer.state_transition_analyze(dFCM_i=dFCM_i, dFCM_j=dFCM_j,  \
        state_match_dict=state_match[session]['method_pairs'][measure_1.measure_name][measure_2.measure_name], \
        TRs=TRs, \
        subject=subj_id, \
        session=session, \
        matching_method='score', 
        verb=True \
    ) # matching_method = score / FCS / transition

#########################################################################################


"""The :mod:`pydfc.dfc_methods` contains dFC methods objects."""

from .base_dfc_method import BaseDFCMethod
from .cap import CAP
from .sliding_window_clustr import SLIDING_WINDOW_CLUSTR
from .discrete_hmm import HMM_DISC
from .continuous_hmm import HMM_CONT
from .sliding_window import SLIDING_WINDOW
from .time_freq import TIME_FREQ
from .windowless import WINDOWLESS

__all__ = ['BaseDFCMethod', 
           'CAP', 'SLIDING_WINDOW_CLUSTR', 
           'HMM_CONT', 'HMM_DISC', 
           'SLIDING_WINDOW', 'TIME_FREQ', 
           'WINDOWLESS' 
           ]




import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
from pathlib import Path
import os

# sys.path.append('./BIC_codes/')
from functions.dFC_funcs import *
from functions.post_analysis_funcs import *

print(' VISUALIZATION IN PROGRESS ... ')

################################# LOAD RESULTS #################################

output_root = './../../../../RESULTs/methods_implementation/server/methods_implementation/final_figures/'
# output_root = './output/'
save_image = True

# the dictionary that collects all RESULTS
ALL_RESULTS = np.load(output_root+'ALL_RESULTS.npy', allow_pickle='True').item()

################################ common variables #################################

node_networks = ALL_RESULTS['node_networks']

cluster_colors_dict = {
    'forestgreen': ['SlidingWindow', 'Time-Freq'],
    'coral': ['Clustering', 'ContinuousHMM', 'DiscreteHMM'],
    'crimson': ['CAP', 'Windowless'],
}

extra_colors = ['khaki', 'orchid', 'aquamarine']

################################# dFC SAMPLES #################################

RESULTS = ALL_RESULTS['dFC_sample']

plot_sample_dFC(D=RESULTS, x='samples',
    title='dFC_samples',
    cmap='seismic',
    normalize=False,
    disp_diag=False,
    save_image=save_image, output_root=output_root+'dFC_sample/', 
    fix_lim=False, center_0=True, 
    node_networks=node_networks, segmented=False 
)

plot_sample_dFC(D=RESULTS, x='samples_ranked',
    title='dFC_samples_ranked',
    cmap='plasma',
    normalize=False,
    disp_diag=False,
    save_image=save_image, output_root=output_root+'dFC_sample/', 
    fix_lim=False, center_0=False, 
    node_networks=node_networks, segmented=False 
)

################################# FCS visualization #################################

for measure in ALL_RESULTS['measure_lst']:

    visualize_FCS(
        measure,
        normalize=True, fix_lim=False, 
        save_image=save_image, output_root=output_root+'FCS/'
    )

################################# RSNs visualization #################################

measure = ALL_RESULTS['measure_lst'][0]

# find ROI count in each RSN
RSNs = np.unique(node_networks)
num_ROIs = {}
for RSN in RSNs:
    num_ROIs[RSN] = np.sum([1 for node in node_networks if node==RSN])

# write to a txt file
folder = f"{output_root}RSNs/"
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/ROI_count_in_RSNs.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')
for RSN in num_ROIs:
    text_file.write(RSN + ' : ' + str(num_ROIs[RSN])+'\n')
text_file.close()

plot_rois(
    node_networks, 
    measure.TS_info['nodes_locs'], 
    save_image=save_image, 
    output_root=f"{output_root}RSNs/"
)

################################# dFC values distributions #################################
dFC_dist_plot = True

if dFC_dist_plot:
    
    RESULTS = ALL_RESULTS['dFC_values_dist']

    ############ VISUALIZE ############

    joint_dist_plot(data=RESULTS,
        title='dFC values distributions',
        save_image=save_image, output_root=output_root+'indiv_prop/'
    )

################################# dFC Similarity #################################

RESULTS = ALL_RESULTS['dFC_similarity_overall']['spearman']['session_Rest1_LR']['overall_stat']

folder = output_root+'dFC_similarity'
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/overall_stat.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')
text_file.write('Average similarity of all pairs (E_E) = '+ str(RESULTS['E_E'])+'\n')
text_file.write('Variance of average similarity across pairs (VAR_E) = '+ str(RESULTS['VAR_E'])+ 
                ' and std = '+ str(np.sqrt(RESULTS['VAR_E']))+'\n')
text_file.write('Average variance of similarity across subjects (E_VAR) = '+ str(RESULTS['E_VAR'])+'\n')
text_file.write('ratio of variances (VAR_E / E_VAR) = '+ str(RESULTS['VAR_E']/RESULTS['E_VAR'])+'\n')
text_file.close()

################# whole subject #################
'''
    - corr((all dFC timepoints of one Subj using method_i), (all dFC timepoints of one Subj using method_j)) -> avg(corr) and var(corr)
    - spearman correlation, pearson correlation, Mutual Information (MI), Euclidean Distance
    - dendogram based on avg(corr)
'''

for metric in ALL_RESULTS['dFC_similarity_overall'] :

    RESULTS = ALL_RESULTS['dFC_similarity_overall'][metric]
    filters_lst = [filter for filter in RESULTS]
    keys_lst = [key for key in RESULTS[filters_lst[0]]]

    ############ VISUALIZE ############
    label_dict = {
        'session_Rest1_LR': 'session Rest1 LR',
        'session_Rest1_RL': 'session Rest1 RL',
        'session_Rest2_LR': 'session Rest2 LR',
        'session_Rest2_RL': 'session Rest2 RL'
    }
    for key in keys_lst:
        if key=='name_lst' or key=='sim_distribution' or key=='overall_stat':
            continue
        visualize_sim_mat(
            RESULTS, 
            mat_key=key, title=metric+' '+key, 
            name_lst_key='name_lst', 
            label_dict=label_dict,
            cmap='viridis',
            save_image=save_image, output_root=output_root+'dFC_similarity/'+metric+'/'
        )

    #### similarity values distribution ####
    label_dict = {
        'sim': 'similarity'
    }
    for filter in ['session_Rest1_LR']:
        pairwise_cat_plots(RESULTS[filter]['sim_distribution'], y='sim',
            title=metric+' total similarity distributions '+filter,
            label_dict=label_dict,
            save_image=save_image, output_root=output_root+'dFC_similarity/'+metric+'/'
        )

    ############ Hierarchical Clustering ############
    for filter in ['session_Rest1_LR']:
        dist_mat = corr2distance(RESULTS[filter]['avg_mat'], metric=metric)
        Z = distance2Z(dist_mat, method='ward')

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(Z=Z, labels=RESULTS[filter]['name_lst'],
                                    cluster_colors_dict=cluster_colors_dict, 
                                    extra_colors=extra_colors,
                    )
        
        dist_mat_dendo(Z=Z, labels=RESULTS[filter]['name_lst'], 
            link_colors=link_cols,
            plot_threshold=True,
            title='Hierarchical Clustering of Methods ' + filter+' using '+metric, 
            save_image=save_image, output_root=output_root+'dFC_similarity/'+metric+'/'
        )

################# Hierarchical Clstr with Confidence Interval #################
'''
    - plot the most frequent hierclstr structures across subjects
'''
min_freq = 10
Z_lst = ALL_RESULTS['Hierclstr_CI']
measures_lst = [measure.measure_name for measure in ALL_RESULTS['measure_lst']]

# cluster Zs to find common structures
Z_clstrs = cluster_Z(Z_lst, num_leaf=len(measures_lst))

num_clstrs = len([key for key in Z_clstrs])
num_clstrs_included = 0
for key in Z_clstrs:
    Z_clstrs[key]['distance_lst'] = np.array(Z_clstrs[key]['distance_lst'])
    if Z_clstrs[key]['freq'] > min_freq:
        num_clstrs_included += 1
        n = Z_clstrs[key]['distance_lst'].shape[0]
        avg_distances = np.mean(Z_clstrs[key]['distance_lst'], axis=0)
        std_distances = np.std(Z_clstrs[key]['distance_lst'], axis=0)

        # reconstruct avg Z
        Z = list()
        for i, tree in enumerate(Z_clstrs[key]['Z']):
            Z.append([tree[0], tree[1], avg_distances[i], tree[3]])
        Z = np.array(Z)

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(Z=Z, labels=measures_lst,
                                    cluster_colors_dict=cluster_colors_dict, 
                                    extra_colors=extra_colors,
                    )
        
        dist_mat_dendo(Z, labels=measures_lst, 
            link_colors=link_cols,
            plot_threshold=True,
            distances_CI=std_distances,
            title='Hierclstr clstr '+str(key)+'_'+str(n)+'subjects',
            save_image=save_image, output_root=output_root+'hierclstr_CI/',
        )  

# write to a txt file
folder = output_root+'hierclstr_CI'
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/hierclstr.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')
text_file.write('Number of structures found across subjects = '+ str(num_clstrs)+'\n')
text_file.write('The min number of subjects to be included = '+ str(min_freq)+'\n')
text_file.write('Number of structures included = '+ str(num_clstrs_included)+'\n')
text_file.close()
        
################# session ANOVA #################

if 'session_ANOVA' in ALL_RESULTS:

    RESULTS = ALL_RESULTS['session_ANOVA'] 
    measure_lst = [measure.measure_name for measure in ALL_RESULTS['measure_lst']]

    data = {
        'day p-values': {
                            'p_values':np.nan*np.ones((len(measure_lst), len(measure_lst))), 
                            'measure_lst':measure_lst
        }, 
        'direction p-values': {
                            'p_values':np.nan*np.ones((len(measure_lst), len(measure_lst))), 
                            'measure_lst':measure_lst
        }, 
    }

    for i, measure_i in enumerate(RESULTS):
        for j, measure_j in enumerate(RESULTS[measure_i]):
            result = RESULTS[measure_i][measure_j]

            index_i = measure_lst.index(unzip_name(measure_i))
            index_j = measure_lst.index(unzip_name(measure_j))

            data['day p-values']['p_values'][index_i, index_j] = result['PR(>F)'][0]
            data['day p-values']['p_values'][index_j, index_i] = result['PR(>F)'][0]
            data['direction p-values']['p_values'][index_i, index_j] = result['PR(>F)'][1]
            data['direction p-values']['p_values'][index_j, index_i] = result['PR(>F)'][1]

    ############ VISUALIZE ############

    visualize_sim_mat(data, mat_key='p_values', title='session_ANOVA_test', 
        name_lst_key='measure_lst', 
        annot=True, fmt=2, 
        show_diag=False, show_sig=True, no_color=True,
        save_image=save_image, output_root=output_root+'dFC_similarity/'
    )

################# feature-based #################
'''
    - spatial
    - temporal
    - inter-time-correlation
    - inter-connection-correlation
    - dFC-avg
    - dFC-var
'''

for feature2extract in ALL_RESULTS['dFC_similarity_feature_based']:

    RESULTS = ALL_RESULTS['dFC_similarity_feature_based'][feature2extract]
    filters_lst = [filter for filter in RESULTS]
    keys_lst = [key for key in RESULTS[filters_lst[0]]]

    ############ VISUALIZE ############
    label_dict = {
        'default_values': 'session Rest1 LR',
        'session_Rest1_RL': 'session Rest1 RL',
        'session_Rest2_LR': 'session Rest2 LR',
        'session_Rest2_RL': 'session Rest2 RL'
    }
    for key in keys_lst:
        if key=='name_lst':
            continue
        visualize_sim_mat(
            RESULTS, mat_key=key, title=feature2extract+' '+key, 
            name_lst_key='name_lst',
            label_dict=label_dict,
            cmap='viridis',
            save_image=save_image, output_root=output_root+'feature_based/'+feature2extract+'/'
        )
    ############ Hierarchical Clustering ############
    for filter in ['default_values']:
        dist_mat = corr2distance(RESULTS[filter]['avg_mat'], metric='spearman')
        Z = distance2Z(dist_mat, method='ward')

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(Z=Z, labels=RESULTS[filter]['name_lst'],
                                    cluster_colors_dict=cluster_colors_dict, 
                                    extra_colors=extra_colors,
                    )
        
        dist_mat_dendo(Z=Z, labels=RESULTS[filter]['name_lst'], 
            link_colors=link_cols,
            plot_threshold=True,
            title='Hierarchical Clustering of Methods ' + filter+' using '+feature2extract, 
            save_image=save_image, output_root=output_root+'feature_based/'+feature2extract+'/'
        )

############ Spatial vs. Temporal Scatter plot ############
    
scatter_data = ALL_RESULTS['spatial_vs_temporal_similarity']

############ visualization ############
label_dict = {
    'spatial': 'spatial similarity',
    'temporal': 'temporal similarity'
}
scatter_plot(
    data=scatter_data, x='temporal', y='spatial', 
    labels='labels', title='spatial similarity vs temporal similarity',
    label_dict=label_dict,
    c=0.3,
    equal_axis_lim=True, show_x_equal_y=True,
    save_image=save_image, output_root=output_root+'variation/spatialVsTemporal/'
)
    
################# graph-based #################
'''
    - spatial
    - temporal-avg
    - ECM, shortest_path, degree, clustering_coef
'''

###### spatial #####

for graph_property in ALL_RESULTS['dFC_similarity_graph']['spatial']:

    RESULTS = ALL_RESULTS['dFC_similarity_graph']['spatial'][graph_property]
    filters_lst = [filter for filter in RESULTS]
    keys_lst = [key for key in RESULTS[filters_lst[0]]]

    ############ VISUALIZE ############
    label_dict = {
        'default_values': 'session Rest1 LR',
        'session_Rest1_RL': 'session Rest1 RL',
        'session_Rest2_LR': 'session Rest2 LR',
        'session_Rest2_RL': 'session Rest2 RL'
    }
    for key in keys_lst:
        if key=='name_lst':
            continue
        visualize_sim_mat(
            RESULTS, mat_key=key, title='spatial '+graph_property+' '+key, 
            name_lst_key='name_lst', 
            label_dict=label_dict,
            cmap='viridis',
            save_image=save_image, output_root=output_root+'graph_based/'+graph_property+'/'
        )
    ############ Hierarchical Clustering ############
    for filter in ['default_values']:
        dist_mat = corr2distance(RESULTS[filter]['avg_mat'], metric='spearman')
        Z = distance2Z(dist_mat, method='ward')

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(Z=Z, labels=RESULTS[filter]['name_lst'],
                                    cluster_colors_dict=cluster_colors_dict, 
                                    extra_colors=extra_colors,
                    )
        
        dist_mat_dendo(Z=Z, labels=RESULTS[filter]['name_lst'], 
            link_colors=link_cols,
            plot_threshold=True,
            title='Hierarchical Clustering of Methods ' + filter+' using '+ 'spatial '+ graph_property, 
            save_image=save_image, output_root=output_root+'graph_based/'+graph_property+'/'
        )

################################# TSNE visualization #################################

'''
    - visualize subjects dFC obtained by different methods in a 2D space
        considering their distances
'''

RESULTS = ALL_RESULTS['TSNE']
sample_measure_lst = [sample[sample.find('_')+1:] for sample in RESULTS['sample_lst']]

measures_lst = list(set(sample_measure_lst))
measures_lst.sort()
colors_lst = ['black', 'green', 'orange', 'red', 'dodgerblue', 'grey', 'violet']
color_dict = {}
for i, measure in enumerate(measures_lst):
    color_dict[measure] = colors_lst[i]

# using overall corr
dist_mat = corr2distance(RESULTS['corr'], metric='spearman')

plot_TSNE(
    dist_mat, 
    sample_measure_lst,
    color_dict,
    projection='2d',
    title='TSNE overall corr',
    save_image=save_image, output_root=output_root+'TSNE/',
)

for n_components in RESULTS['X_red_corr']:
    dist_mat = corr2distance(RESULTS['X_red_corr'][n_components], metric='spearman')

    plot_TSNE(
        dist_mat, 
        sample_measure_lst,
        color_dict,
        projection='2d',
        title='TSNE '+str(n_components),
        save_image=save_image, output_root=output_root+'TSNE/',
    )

################################# inter_subject similarity #################################

'''
    - returns correspondence of inter-subject relation between results of dFC 
        measures in each session
    - dendogram based on inter-subject similarity
'''

for subj_lvl_feature in ALL_RESULTS['subj_clustring']:

    RESULTS = ALL_RESULTS['subj_clustring'][subj_lvl_feature] 

    ############ VISUALIZE ############
    label_dict = {
        'session_Rest1_LR': 'session Rest1 LR',
        'session_Rest1_RL': 'session Rest1 RL',
        'session_Rest2_LR': 'session Rest2 LR',
        'session_Rest2_RL': 'session Rest2 RL'
    }
    for key in RESULTS:
        annot = True
        # set diag values to 0
        for session in RESULTS[key]:
            np.fill_diagonal(RESULTS[key][session]['sim_mat'], 0)
        visualize_sim_mat(
            RESULTS[key], mat_key='sim_mat', title='inter-subject-corr similarity '+key+ ' based on '+subj_lvl_feature, 
            name_lst_key='name_lst',
            label_dict=label_dict,
            cmap='viridis',
            annot=annot,
            save_image=save_image, output_root=output_root+'inter_subject/'+subj_lvl_feature+'/'
        )
    ############ Hierarchical Clustering ############
    for session in RESULTS['across_method']:
        dist_mat = corr2distance(RESULTS['across_method'][session]['sim_mat'], metric='spearman')
        Z = distance2Z(dist_mat, method='ward')

        # find link colors to fix the color of each cluster
        link_cols = find_link_colors(Z=Z, labels=RESULTS['across_method'][session]['name_lst'],
                                    cluster_colors_dict=cluster_colors_dict, 
                                    extra_colors=extra_colors,
                    )
        
        dist_mat_dendo(Z=Z, labels=RESULTS['across_method'][session]['name_lst'], 
            link_colors=link_cols,
            plot_threshold=True,
            title='Hierarchical Clustering of Methods ' + session +' using inter-subject similarity based on '+subj_lvl_feature, 
            save_image=save_image, output_root=output_root+'inter_subject/'+subj_lvl_feature+'/'
        )

################################# dFC var #################################

'''
    - avg(variance/fluctuations of dFC in one Subj)
    - rank normed
'''

RESULTS = ALL_RESULTS['dFC_var']

visualize_conn_mat_dict(RESULTS['avg_dFC_var'], node_networks=node_networks, 
            title='avg dFC var', center_0=False,
            fix_lim=False, disp_diag=True, cmap='plasma', normalize=False, 
            save_image=save_image, output_root=output_root+'dFC_var/'
)

visualize_conn_mat_dict(RESULTS['avg_dFC_var'], node_networks=node_networks, segmented=True,
            title='segmented avg dFC var', center_0=False,
            fix_lim=False, disp_diag=True, cmap='plasma', normalize=False, 
            save_image=save_image, output_root=output_root+'dFC_var/'
)

visualize_conn_mat_dict(RESULTS['var_dFC_var'], node_networks=node_networks, 
            title='var of dFC var', center_0=False,
            fix_lim=False, disp_diag=True, cmap='plasma', normalize=False, 
            save_image=save_image, output_root=output_root+'dFC_var/'
)

################################# dFC avg #################################

'''
    - avg(avg of dFC -static FC- in one Subj)
    - rank normed
'''

RESULTS = ALL_RESULTS['dFC_avg'] 

visualize_conn_mat_dict(RESULTS, node_networks=node_networks, 
        title='dFC avg', center_0=False,
        fix_lim=False, disp_diag=False, cmap='plasma', normalize=False,
        save_image=save_image, output_root=output_root+'dFC_avg/'
)

visualize_conn_mat_dict(RESULTS, node_networks=node_networks, segmented=True,
        title='segmented dFC avg', center_0=False,
        fix_lim=False, disp_diag=False, cmap='plasma', normalize=False,
        save_image=save_image, output_root=output_root+'dFC_avg/'
)

################################# Across Func Conn total Correlation #################################

'''
    - spearman_corr((dFConnection(node_i, node_j) timecourse using method m), (dFConnection(node_i, node_j) timecourse using method n))
'''

RESULTS = ALL_RESULTS['across_func_conns']

############ VISUALIZE ############

visualize_conn_mat_2D_dict(RESULTS, node_networks=node_networks, 
    title='across func conn total spearman corr', fix_lim=False, 
    disp_diag=False, cmap='seismic', normalize=False, center_0=True,
    save_image=save_image, output_root=output_root+'across_func_conn/'
)

visualize_conn_mat_2D_dict(RESULTS, node_networks=node_networks, segmented=True,
    title='segmented across func conn total spearman corr', fix_lim=False, 
    disp_diag=False, cmap='seismic', normalize=False, center_0=True,
    save_image=save_image, output_root=output_root+'across_func_conn/'
)

visualize_conn_mat_2D_dict(RESULTS, node_networks=node_networks, 
    title='across func conn total spearman corr normalized', fix_lim=False, 
    disp_diag=False, cmap='seismic', normalize=True, center_0=True,
    save_image=save_image, output_root=output_root+'across_func_conn/'
)

visualize_conn_mat_2D_dict(RESULTS, node_networks=node_networks, segmented=True,
    title='segmented across func conn total spearman corr normalized', fix_lim=False, 
    disp_diag=False, cmap='seismic', normalize=True, center_0=True,
    save_image=save_image, output_root=output_root+'across_func_conn/'
)

################################# High Variation Regions #################################
'''
    - high variation regions over methods and over time.
'''

RESULTS = ALL_RESULTS['var_across_func_conns'] 
RATIO = ALL_RESULTS['var_across_func_conns_ratio'] 
scatter_data = ALL_RESULTS['var_method_vs_time_across_func_conns_scatter'] 

############ method var / time var ############

ratio = list()
ratio_std = list()
for i, sample in enumerate(scatter_data['var_method']):
    ratio.append(scatter_data['var_method'][i]/scatter_data['var_time'][i])
    ratio_std.append(np.sqrt(scatter_data['var_method'][i])/np.sqrt(scatter_data['var_time'][i]))
ratio = np.mean(np.array(ratio))
ratio_std = np.mean(np.array(ratio_std))    
avg_var_time = np.mean(scatter_data['var_time'])
avg_var_method = np.mean(scatter_data['var_method'])

# write to a txt file
folder = output_root+'variation/timeVsMethod'
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/var_ratio.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')
text_file.write('Average of var method / var time ratio = '+ str(ratio)+'\n')
text_file.write('Average of std method / std time ratio = '+ str(ratio_std)+'\n')
text_file.write('Average var method = '+ str(avg_var_method)+'\n')
text_file.write('Average var time = '+ str(avg_var_time)+'\n')
text_file.write('Ratio of avg var method / avg var time = '+ str(avg_var_method/avg_var_time)+'\n')
text_file.close()

############ VISUALIZE ############

label_dict = {
    'var_method': 'variation over method',
    'var_time': 'variation over time'
}
scatter_plot(
    data=scatter_data, x='var_time', y='var_method', 
    title='var method vs time across func conns',
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True, show_x_equal_y=True,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

label_dict = {
    'var across method / time - 1': 'var over method / time - 1',
}
visualize_conn_mat_dict(RATIO, node_networks=node_networks, segmented=False,
    title='ratio of high variation regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=False, cmap='seismic', center_0=True,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

visualize_conn_mat_dict(RATIO, node_networks=node_networks, segmented=True,
    title='segmented ratio of high variation regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=False, cmap='seismic', center_0=True,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

label_dict = {
    'var_over_method': 'variation over method',
    'var_over_time': 'variation over time'
}

visualize_conn_mat_dict(RESULTS, node_networks=node_networks, 
    title='variation across regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

# func conn segmented
visualize_conn_mat_dict(RESULTS, node_networks=node_networks, segmented=True,
    title='segmented high variation regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

RESULTS = ALL_RESULTS['high_var_func_conns'] 

label_dict = {
    'var_over_method': 'variation over method',
    'var_over_time': 'variation over time'
}
visualize_conn_mat_dict(RESULTS, node_networks=node_networks, 
    title='high variation regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

################################# Variation Value Comparison #################################
'''
    - compare variation over methods with variation over time.
'''

RESULTS = ALL_RESULTS['var_comparison']
scatter_data_across_func_conn = ALL_RESULTS['var_method_vs_time_method_pairs_across_func_conns'] 
scatter_data = ALL_RESULTS['var_method_vs_time_method_pairs']

############ VISUALIZE ############

visualize_sim_mat(
    RESULTS, mat_key='sim_mat', title='variation in different dimensions', 
    name_lst_key='name_lst', 
    cmap='viridis',
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

label_dict = {
    'var_method': 'variation over methods',
    'var_time': 'variation over time'
}
pairwise_scatter_plots(
    data=scatter_data_across_func_conn, x='var_time', y='var_method', 
    title='var method vs time across func conns across methods pairs', 
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True, show_x_equal_y=True,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

label_dict = {
    'var_method': 'variation over method',
    'var_time': 'variation over time'
}
scatter_plot(
    data=scatter_data, x='var_time', y='var_method', 
    labels='labels', title='var method vs time',
    label_dict=label_dict,
    c=0.5,
    equal_axis_lim=True, show_x_equal_y=True,
    save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
)

################################# Var method vs. Time Clstrwise #################################
'''
    - compute var over method and time within each group of methods
'''
scatter_data = ALL_RESULTS['var_method_vs_time_clstrwise']['scatter_data']
clstrs_dict = ALL_RESULTS['var_method_vs_time_clstrwise']['clstrs_dict']

# write to a txt file
folder = output_root+'variation'
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/var_clstrwise_ratio.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')

for clstr in scatter_data:

    ratio = list()
    for i, sample in enumerate(scatter_data[clstr]['var_method']):
        ratio.append(scatter_data[clstr]['var_method'][i]/scatter_data[clstr]['var_time'][i])

    ratio = np.mean(np.array(ratio))
    avg_var_time = np.mean(scatter_data[clstr]['var_time'])
    avg_var_method = np.mean(scatter_data[clstr]['var_method'])
    
    text_file.write(clstr+'\n')
    text_file.write(' '.join(clstrs_dict[clstr])+'\n')
    text_file.write('Average of var method / var time ratio = '+ str(ratio)+'\n')
    text_file.write('Average var method = '+ str(avg_var_method)+'\n')
    text_file.write('Average var time = '+ str(avg_var_time)+'\n')
    text_file.write('Ratio of avg var method / avg var time = '+ str(avg_var_method/avg_var_time)+'\n')

    label_dict = {
        'var_method': 'variation over method',
        'var_time': 'variation over time'
    }
    scatter_plot(
        data=scatter_data[clstr], x='var_time', y='var_method', 
        title='var method vs time clusterwise across func conns ' + clstr,
        label_dict=label_dict,
        hist=True,
        equal_axis_lim=True, show_x_equal_y=True,
        save_image=save_image, output_root=output_root+'variation/timeVsMethod/'
    )
text_file.close()

################################# Inter-subject Variation #################################
'''
    - compute var over method and subject across all func conns 
'''

RESULTS = ALL_RESULTS['inter-subj_var_across_func_conns'] 
RATIO = ALL_RESULTS['inter-subj_var_across_func_conns_ratio'] 
scatter_data = ALL_RESULTS['var_method_vs_subj_across_func_conns_scatter'] 

############ method var / subject var ############

ratio = list()
ratio_std = list()
for i, sample in enumerate(scatter_data['var_method']):
    ratio.append(scatter_data['var_method'][i]/scatter_data['var_subj'][i])
    ratio_std.append(np.sqrt(scatter_data['var_method'][i])/np.sqrt(scatter_data['var_subj'][i]))
ratio = np.mean(np.array(ratio))
ratio_std = np.mean(np.array(ratio_std))    
avg_var_subj = np.mean(scatter_data['var_subj'])
avg_var_method = np.mean(scatter_data['var_method'])

# write to a txt file
folder = output_root+'variation/subjVsMethod'
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/subj_var_ratio.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')
text_file.write('Average of var method / var subj ratio = '+ str(ratio)+'\n')
text_file.write('Average of std method / std subj ratio = '+ str(ratio_std)+'\n')
text_file.write('Average var method = '+ str(avg_var_method)+'\n')
text_file.write('Average var subj = '+ str(avg_var_subj)+'\n')
text_file.write('Ratio of avg var method / avg var subj = '+ str(avg_var_method/avg_var_subj)+'\n')
text_file.close()

############ VISUALIZE ############

label_dict = {
    'var_method': 'variation over method',
    'var_subj': 'variation over subject'
}
scatter_plot(
    data=scatter_data, x='var_subj', y='var_method', 
    title='var method vs subj across func conns',
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True, show_x_equal_y=True,
    save_image=save_image, output_root=output_root+'variation/subjVsMethod/'
)

label_dict = {
    'var across method / subj - 1': 'var over method / subj - 1',
}
visualize_conn_mat_dict(RATIO, node_networks=node_networks, segmented=False,
    title='ratio of method subj var regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=False, cmap='seismic', center_0=True,
    save_image=save_image, output_root=output_root+'variation/subjVsMethod/'
)

visualize_conn_mat_dict(RATIO, node_networks=node_networks, segmented=True,
    title='segmented ratio of method subj var regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=False, cmap='seismic', center_0=True,
    save_image=save_image, output_root=output_root+'variation/subjVsMethod/'
)

label_dict = {
    'var_over_method': 'variation over method',
    'var_over_subj': 'variation over subject',
}

visualize_conn_mat_dict(RESULTS, node_networks=node_networks, 
    title='inter-subj variation across regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/subjVsMethod/'
)

# func conn segmented
visualize_conn_mat_dict(RESULTS, node_networks=node_networks, segmented=True,
    title='segmented inter-subj variation across regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/subjVsMethod/'
)

################################# Overall Variability #################################
'''
    - compute var over method and overall across all func conns 
'''

RESULTS = ALL_RESULTS['var_all_across_func_conns'] 
RATIO = ALL_RESULTS['var_all_across_func_conns_ratio'] 
scatter_data = ALL_RESULTS['var_method_vs_all_across_func_conns_scatter'] 

############ method var / subject var ############

ratio = list()
ratio_std = list()
for i, sample in enumerate(scatter_data['var_method']):
    ratio.append(scatter_data['var_method'][i]/scatter_data['var_all'][i])
    ratio_std.append(np.sqrt(scatter_data['var_method'][i])/np.sqrt(scatter_data['var_all'][i]))
ratio = np.mean(np.array(ratio))
ratio_std = np.mean(np.array(ratio_std))    
avg_var_all = np.mean(scatter_data['var_all'])
avg_var_method = np.mean(scatter_data['var_method'])

# write to a txt file
folder = output_root+'variation/allVsMethod'
if not os.path.exists(folder):
    os.makedirs(folder)
filename = Path(folder+'/all_var_ratio.txt')
filename.touch(exist_ok=True)
text_file = open(filename, 'wt')
text_file.write('Average of var method / var all ratio = '+ str(ratio)+'\n')
text_file.write('Average of std method / std all ratio = '+ str(ratio_std)+'\n')
text_file.write('Average var method = '+ str(avg_var_method)+'\n')
text_file.write('Average var subj = '+ str(avg_var_all)+'\n')
text_file.write('Ratio of avg var method / avg var subj = '+ str(avg_var_method/avg_var_all)+'\n')
text_file.close()

############ VISUALIZE ############

label_dict = {
    'var_method': 'variation over method',
    'var_all': 'variation over all'
}
scatter_plot(
    data=scatter_data, x='var_all', y='var_method', 
    title='var method vs all across func conns',
    label_dict=label_dict,
    hist=True,
    equal_axis_lim=True, show_x_equal_y=True,
    save_image=save_image, output_root=output_root+'variation/allVsMethod/'
)

label_dict = {
    'var across method / all - 1': 'var over method / all - 1',
}
visualize_conn_mat_dict(RATIO, node_networks=node_networks, segmented=False,
    title='ratio of method all var regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=False, cmap='seismic', center_0=True,
    save_image=save_image, output_root=output_root+'variation/allVsMethod/'
)

visualize_conn_mat_dict(RATIO, node_networks=node_networks, segmented=True,
    title='segmented ratio of method all var regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=False, cmap='seismic', center_0=True,
    save_image=save_image, output_root=output_root+'variation/allVsMethod/'
)

label_dict = {
    'var_over_method': 'variation over method',
    'var_over_all': 'variation over all',
}

visualize_conn_mat_dict(RESULTS, node_networks=node_networks, 
    title='all variation across regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/allVsMethod/'
)

# func conn segmented
visualize_conn_mat_dict(RESULTS, node_networks=node_networks, segmented=True,
    title='segmented all variation across regions', fix_lim=False,  
    label_dict=label_dict,
    disp_diag=True, cmap='plasma', center_0=False,
    save_image=save_image, output_root=output_root+'variation/allVsMethod/'
)

################################# Randomization Tests #################################

metric = 'spearman'

'''
find the similarity between the dFC obtained by each method 
and a dFC which is a constant sequence of mean of all methods dFC
'''

########### Similarity with static FC ###########

RESULTS = ALL_RESULTS['randomization']['sim_with_static_FC']

############ VISUALIZE ############
label_dict = {
    'sim': 'similarity',
    'dFC_method': 'dFC assessment method'
}
cat_plot(data=RESULTS, x='dFC_method', y='sim', 
        kind='box',
        title='similarity with constant static FC',
        label_dict=label_dict,
        save_image=save_image, output_root=output_root+'randomization/'
        )

########### Shuffled ###########

'''
find the similarity between the dFC obtained by each method 
but with randomized temporal/spatial/all order
'''

RESULTS = ALL_RESULTS['randomization']['shuffled']['similarity_dict']
modes_lst = ALL_RESULTS['randomization']['shuffled']['modes_lst']

############ VISUALIZE ############
label_dict = {
    'all_shuffle_sim': 'similarity',
    'spatial_shuffle_sim': 'similarity',
    'temporal_shuffle_sim': 'similarity',
}
for mode in modes_lst:
    key = mode+'_shuffle_sim'
    pairwise_cat_plots(RESULTS, y=key, z='actual_sim',
        title=mode+' shuffled similarity',
        label_dict=label_dict,
        save_image=save_image, output_root=output_root+'randomization/'
        )

########### Random state time course ###########

'''
find the similarity between dFC matrices created using
a random state time course but real FC patterns obtained 
by each method; for SW and TF all FC patterns of each 
subject are used
'''

RESULTS = ALL_RESULTS['randomization']['random_state_TC'] 

############ VISUALIZE ############
label_dict = {
    'sim': 'similarity',
}
pairwise_cat_plots(RESULTS, y='sim', z='actual_sim',
    title='random state time course dFC',
    label_dict=label_dict,
    save_image=save_image, output_root=output_root+'randomization/'
)

################################# inter-state spatial similarity #################################

key_name = 'inter-state_spatial_sim'
RESULTS = ALL_RESULTS[key_name]

############ VISUALIZE ############
label_dict = {
    'sim': 'inter-state spatial similarity',
    'dFC_method': 'dFC assessment method'
}
cat_plot(data=RESULTS, x='dFC_method', y='sim', 
    kind='box',
    scale_dist=False,
    title=key_name,
    label_dict=label_dict,
    save_image=save_image, output_root=output_root+'indiv_prop/'
    )

################################# SIMILARITY OF ADJACENT TIME POINTS #################################

key_name = 'similarity of adjacent time points'
RESULTS = ALL_RESULTS['adjacent_time_points']

############ VISUALIZE ############
label_dict = {
    'dFC_method': 'dFC assessment method'
}
cat_plot(data=RESULTS, x='dFC_method', y=key_name, 
    kind='box',
    title=key_name,
    label_dict=label_dict,
    save_image=save_image, output_root=output_root+'indiv_prop/'
    )

################################# TRANSITION FREQUENCY #################################
'''
 - plot normalized transition frequency
'''

key_name = 'trans_freq'
RESULTS = ALL_RESULTS['transition_freq'] 

############ VISUALIZE ############
label_dict = {
    'trans_freq': 'transition frequency',
    'dFC_method': 'dFC assessment method'
}
cat_plot(data=RESULTS, x='dFC_method', y=key_name, 
    kind='box',
    title=key_name,
    label_dict=label_dict,
    save_image=save_image, output_root=output_root+'indiv_prop/'
    )

################################# DWELL TIME #################################
'''
 - plot normalized (not downsampled) dwell times
 - the dwell times are not averaged wihtin each subj, they include all individual 
 dwells
'''

key_name = 'dwell_time'
RESULTS = ALL_RESULTS['dwell_time'] 

############ VISUALIZE ############
label_dict = {
    'dwell_time': 'dwell time',
    'dFC_method': 'dFC assessment method'
}
cat_plot(data=RESULTS, x='dFC_method', y=key_name, 
    kind='box',
    scale_dist=True,
    title=key_name,
    label_dict=label_dict,
    y_lim=(0, 0.2),
    save_image=save_image, output_root=output_root+'indiv_prop/'
    )

################################# TIME RECORD #################################

RESULTS = ALL_RESULTS['time_record'] 
filter = 'session_Rest1_LR'

############ VISUALIZE ############
label_dict = {
    'FCS_fit_time (s)':'FCS fitting time (s)',
    'dFC_assess_time (s)':'dFC assessing time (s)',
    'dFC_method': 'dFC assessment method'
}
cat_plot(data=RESULTS, x='dFC_method', y='dFC_assess_time (s)', 
    kind='bar', log=True,
    title='dFC assess time record of ' + filter,
    label_dict=label_dict,
    save_image=save_image, output_root=output_root+'time/'
)

cat_plot(data=RESULTS, x='dFC_method', y='FCS_fit_time (s)', 
    kind='bar', log=True,
    title='FCS fit time record of ' + filter,
    label_dict=label_dict,
    save_image=save_image, output_root=output_root+'time/'
)

if not save_image:
    plt.show()
#################################################################################



import pytest

import nibabel as nb
import numpy as np

from pydfc.data_loader import nifti2timeseries


# @pytest.fixture(scope="session")
# def rest_file(tmp_path_factory):
#     URL = "https://s3.amazonaws.com/openneuro.org/ds002785/derivatives/fmriprep/sub-0001/func/sub-0001_task-restingstate_acq-mb3_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz?versionId=UfCs4xtwIEPDgmb32qFbtMokl_jxLUKr"
#     tmpdir = tmp_path_factory.mktemp("data")
#     file_path = tmpdir / "sub-0001_task-restingstate_acq-mb3_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
#     with httpx.stream("GET", URL) as response:
#         with file_path.open("wb") as f:
#             for chunk in response.iter_bytes():
#                 f.write(chunk)
#
#     return file_path


@pytest.fixture
def simulated_bold_data(tmp_path):
    img = nb.Nifti1Image(np.random.rand(10, 10, 10, 100), np.eye(4))
    img.to_filename(tmp_path / "simulated_bold.nii.gz")
    return tmp_path / "simulated_bold.nii.gz"


def test_load(simulated_bold_data):
    timeseries = nifti2timeseries(
        nifti_file=str(simulated_bold_data),
        n_rois=100,
        Fs=1 / 0.75,
        subj_id="sub-0001",
    )



from pydfc import (
    data_loader,
    MultiAnalysis,
)
import time
import numpy as np
import os
import warnings

warnings.simplefilter('ignore')

os.environ["MKL_NUM_THREADS"] = '16'
os.environ["NUMEXPR_NUM_THREADS"] = '16'
os.environ["OMP_NUM_THREADS"] = '16'

################################# Parameters #################################
# data paths
# main_root = '../../DATA/ds002785/' # for local
main_root = '../../../DATA/task-based/openneuro/ds002785' # for server
roi_root = f"{main_root}/derivatives/ROI_timeseries"
output_root = f"{main_root}/derivatives/fitted_MEASURES"

# for consistency we use 0 for resting state
TASKS = [
    'task-restingstate', 
    'task-anticipation', 
    'task-emomatching', 
    'task-faces', 
    'task-gstroop', 
    'task-workingmemory'
]

job_id = int(os.getenv("SGE_TASK_ID"))
TASK_id = job_id-1 # SGE_TASK_ID starts from 1 not 0
if TASK_id >= len(TASKS):
    print("TASK_id out of TASKS")
    exit()
task = TASKS[TASK_id]

###### MEASUREMENT PARAMETERS ######

# W is in sec

params_methods = { 
    # Sliding Parameters
    'W': 44, 'n_overlap': 1.0, 'sw_method':'pear_corr', 'tapered_window':True, 
    # TIME_FREQ
    'TF_method':'WTC', 
    # CLUSTERING AND DHMM
    'clstr_base_measure':'SlidingWindow', 
    # HMM
    'hmm_iter': 20, 'dhmm_obs_state_ratio': 16/24, 
    # State Parameters
    'n_states': 12, 'n_subj_clstrs': 20, 
    # Parallelization Parameters
    'n_jobs': 2, 'verbose': 0, 'backend': 'loky', 
    # SESSION
    'session': task, 
    # Hyper Parameters
    'normalization': True, 
    'num_subj': None, # None or 216?
    'num_time_point': None,  # None or set?
}

###### HYPER PARAMETERS ALTERNATIVE ######

MEASURES_name_lst = [ 
    'SlidingWindow', 
    'Time-Freq', 
    'CAP', 
    'ContinuousHMM', 
    'Windowless', 
    'Clustering', 
    'DiscreteHMM' 
]

alter_hparams = { 
    # 'session': ['Rest1_RL', 'Rest2_LR', 'Rest2_RL'], 
    # 'n_overlap': [0, 0.25, 0.75, 1], 
    # 'n_states': [6, 16], 
    # # 'normalization': [], 
    # 'num_subj': [50, 100, 200], 
    # 'num_select_nodes': [30, 50, 333], 
    # 'num_time_point': [800, 1000], 
    # 'Fs_ratio': [0.50, 0.75, 1.5], 
    # 'noise_ratio': [1.00, 2.00, 3.00], 
    # 'num_realization': [] 
}

###### MultiAnalysis PARAMETERS ######

params_multi_analysis = { 
    # Parallelization Parameters
    'n_jobs': None, 'verbose': 0, 'backend': 'loky' 
}

################################# LOAD DATA #################################

BOLD = data_loader.load_TS(
        data_root=roi_root, 
        file_name='time_series.npy', 
        SESSIONs=task, 
        subj_id2load=None
)

################################# Visualize BOLD #################################

# for session in BOLD:
#     BOLD.visualize(start_time=0, end_time=2000, nodes_lst=list(range(10)),
#         save_image=False, output_root=None)

################################ Measures of dFC #################################

MA = MultiAnalysis( 
    analysis_name=f"task-based-dFC-ds002785-{task}",
    **params_multi_analysis 
)

MEASURES_lst = MA.measures_initializer(
    MEASURES_name_lst, 
    params_methods, 
    alter_hparams 
)

tic = time.time()
print('Measurement Started ...')

################################# estimate FCS #################################

for MEASURE_id, measure in enumerate(MEASURES_lst):

    print('MEASURE: ' + measure.measure_name)
    print("FCS estimation started...")

    if measure.is_state_based:
        measure.estimate_FCS(time_series=BOLD)
            
    # dFC_analyzer.estimate_group_FCS(time_series_dict=BOLD)
    print("FCS estimation done.")

    # Save
    if not os.path.exists(f"{output_root}/{task}"):
        os.makedirs(f"{output_root}/{task}")
    np.save(f"{output_root}/{task}/MEASURE_{str(MEASURE_id)}.npy", measure)

print('Measurement required %0.3f seconds.' % (time.time() - tic, ))
np.save(f"{output_root}/{task}/multi_analysis.npy", MA)

#################################################################################



# -*- coding: utf-8 -*-
"""
functions for facilitating dynamic functional connectivity analysis

Created on Jun 29 2023
@author: Mohammad Torabi
"""

import numpy as np
from scipy import signal
from copy import deepcopy
import matplotlib.pyplot as plt
import seaborn as sns
from nilearn.plotting import plot_markers
import networkx as nx
from scipy import stats
import os

# np.seterr(invalid='ignore')

# ########## bundled brain graph visualizer ##########


# import panel as pn
# import datashader as ds
# import datashader.transfer_functions as tf
# from datashader.layout import random_layout, circular_layout, forceatlas2_layout
# from datashader.bundling import connect_edges, hammer_bundle
# from datashader import utils
# import holoviews as hv
# from itertools import chain

# import warnings

# warnings.simplefilter('ignore')

################################# Parameters ####################################

fig_dpi = 120
fig_bbox_inches = 'tight'
fig_pad = 0.1
show_title = False
save_fig_format = 'png'

################################# Other Functions ####################################

# test
def zip_name(name):
    # zip measure names
    if 'Clustering' in name:
        new_name = 'SWC' 
    if 'CAP' in name:
        new_name = 'CAP' 
    if 'ContinuousHMM' in name:
        new_name = 'CHMM' 
    if 'Windowless' in name:
        new_name = 'WL' 
    if 'DiscreteHMM' in name:
        new_name = 'DHMM' 
    if 'Time-Freq' in name:
        new_name = 'TF' 
    if 'SlidingWindow' in name:
        new_name = 'SW'
    return new_name

# test
# pear_corr problem
def unzip_name(name):
    # unzip measure names
    if 'SWC' in name:
        new_name = 'Clustering' 
    elif 'CAP' in name:
        new_name = 'CAP' 
    elif 'CHMM' in name:
        new_name = 'ContinuousHMM' 
    elif 'WL' in name:
        new_name = 'Windowless' 
    elif 'DHMM' in name:
        new_name = 'DiscreteHMM' 
    elif 'TF' in name:
        new_name = 'Time-Freq' 
    elif 'SW' in name:
        new_name = 'SlidingWindow'
    return new_name

def find_new_order(old_list, new_list):
    '''
    new_order is a list of indices
    old_list = ['E', 'B', 'A', 'C', 'D']
    new_list = ['A', 'B', 'C', 'D', 'E']
    '''
    new_order = [old_list.index(a) for a in new_list]  
    return new_order

def mat_reorder(A, new_order):
    '''
    new_order must be a list of indices:
    old_list = ['E', 'B', 'A', 'C', 'D']
    new_list = ['A', 'B', 'C', 'D', 'E']
    new_order = find_new_order(old_list, new_list)
    A_sorted is a copy of A
    '''
    assert (
        len(new_order)==A.shape[0]
        and len(new_order)==A.shape[1]
    ), 'dimension mismatch in reordering.'
    A_sorted = deepcopy(A)

    A_sorted = [[A_sorted[i][j] for j in new_order] for i in new_order]
    A_sorted = np.array(A_sorted)
    return A_sorted

# test
def get_subj_ts_dict(time_series_dict, subjs_id):
    subj_ts_dict = {}
    for session in time_series_dict:
        subj_ts_dict[session] = time_series_dict[session].get_subj_ts(subjs_id=subjs_id)
    return subj_ts_dict

# test
def filter_dFC_lst(dFC_lst, **param_dict):
    dFC_lst2check = list()
    for dFC in dFC_lst:
        if dFC.measure.param_match(**param_dict):
            dFC_lst2check.append(dFC) 
    return dFC_lst2check

def SW_downsample(data, Fs, W, n_overlap, tapered_window=False):
    '''
    data = (n_time, ...)
    the time samples will be picked after 
    averaging over a window which slides
    W is in sec
    SWed_data = (n_time_new, ...)
    '''

    SWed_data = list()
    L = data.shape[0]
    # change W to timepoints
    W = int(W * Fs) 
    step = int((1-n_overlap)*W)
    if step == 0:
        step = 1

    window_taper = signal.windows.gaussian(W, std=3*W/22)

    TR_array = list()
    for l in range(0, L-W+1, step):

        ######### creating a rectangel window ############
        window = np.zeros((L))
        window[l:l+W] = 1
        
        ########### tapering the window ##############
        if tapered_window:
            window = signal.convolve(window, window_taper, mode='same') / sum(window_taper)

        # int(l-W/2):int(l+3*W/2) is the nonzero interval after tapering
        SWed_data.append(np.average(data, weights=window, axis=0))
        
        TR_array.append(int((l + (l+W)) / 2) )
        
    
    SWed_data = np.array(SWed_data)
    return SWed_data

def mutual_information(X, Y, N_bins=100):
    """ Mutual information for joint histogram
    https://matthew-brett.github.io/teaching/mutual_information.html#:~:text=Mutual%20information%20is%20a%20measure,signal%20intensity%20in%20the%20first.
    """

    # 2D histogram
    hist_2d, x_edges, y_edges = np.histogram2d(
                                                X,
                                                Y,
                                                bins=N_bins)
    
    # Convert bins counts to probability values
    pxy = hist_2d / float(np.sum(hist_2d))
    px = np.sum(pxy, axis=1) # marginal for x over y
    py = np.sum(pxy, axis=0) # marginal for y over x
    px_py = px[:, None] * py[None, :] # Broadcast to multiply marginals
    # Now we can do the calculation using the pxy, px_py 2D arrays
    nzs = pxy > 0 # Only non-zero pxy values contribute to the sum
    return np.sum(pxy[nzs] * np.log(pxy[nzs] / px_py[nzs]))

def corr2distance(corr_mat, metric):
    '''
    metric can be: MI, euclidean_distance, spearman,
    corr (pearson)
    '''
    if metric=='MI':
        if np.any(corr_mat>1):
            print('MI values cannot be converted to distances.')
        # negative corr will be > 1.0
        dist_mat = 1 - corr_mat
    elif metric=='euclidean_distance':
        dist_mat = corr_mat
    else:
        # negative corr will be > 1.0
        dist_mat = 1 - corr_mat
    # dist_mat must be symmetric
    dist_mat = 0.5*(dist_mat + dist_mat.T)
    # diagonal values of dist_mat must equal exactly zero
    np.fill_diagonal(dist_mat, 0)
    return dist_mat

# test
def normalizeAdjacency(W):
    """
    NormalizeAdjacency: Computes the [0, 1]-normalized adjacency matrix

    Input:

        W (np.array): adjacency matrix

    Output:

        W_norm (np.array): [0, 1] normalized adjacency matrix
    """
    W_norm = W - np.min(W)
    W_norm = np.divide(W_norm, np.max(W_norm))
    return W_norm 

# test
def normalized_euc_dist(x, y):
    # https://stats.stackexchange.com/questions/136232/definition-of-normalized-euclidean-distance#:~:text=The%20normalized%20squared%20euclidean%20distance,not%20related%20to%20Mahalanobis%20distance.

    if np.linalg.norm(x-np.mean(x))**2==0 and np.linalg.norm(y-np.mean(y))**2==0:
        return 0
    return 0.5*((np.linalg.norm((x-np.mean(x)) - (y-np.mean(y)))**2)/(np.linalg.norm(x-np.mean(x))**2 + np.linalg.norm(y-np.mean(y))**2))

def calc_graph_propoerty(A, property, threshold=False, binarize=False):
    """
    calc_graph_propoerty: Computes Graph-based properties 
    of adjacency matrix A
    A is converted to positive before calc
    property:
        - ECM: Computes Eigenvector Centrality Mapping (ECM) 
        - shortest_path
        - degree
        - clustering_coef

    Input:

        A (np.array): adjacency matrix (must be >0)

    Output:

        graph-property (np.array): a vector
    """
    N_edges = 200 # number of edges to keep 
    if property=='shortest_path' or property=='clustering_coef':
        threshold=True

    G = nx.from_numpy_array(np.abs(A)) 
    G.remove_edges_from(nx.selfloop_edges(G))
    # G = G.to_undirected()

    # pruning edges 
    if threshold:
        labels = [d["weight"] for (u, v, d) in G.edges(data=True)]
        labels.sort()
        threshold = labels[-1*N_edges]
        ebunch = [(u, v) for u, v, d in G.edges(data=True) if d['weight']<threshold]
        G.remove_edges_from(ebunch)

    if binarize:
        weight='None'
    else:
        weight='weight'

    if property=='ECM':
        graph_property = nx.eigenvector_centrality(G, weight=weight)
        graph_property = [graph_property[node] for node in graph_property]
        graph_property = np.array(graph_property)
    if property=='shortest_path':
        SHORTEST_PATHS = dict(nx.shortest_path_length(G, weight=weight))
        graph_property = np.zeros((A.shape[0], A.shape[0]))
        for node_i in SHORTEST_PATHS:
            for node_j in SHORTEST_PATHS[node_i]:
                graph_property[node_i, node_j] = SHORTEST_PATHS[node_i][node_j]
        graph_property = graph_property + graph_property.T
        graph_property = dFC_mat2vec(graph_property)
    if property=='degree':
        graph_property = [G.degree(weight=weight)[node] for node in G]
        graph_property = np.array(graph_property)
    if property=='clustering_coef':
        graph_property = nx.clustering(G, weight=weight)
        graph_property = [graph_property[node] for node in graph_property]
        graph_property = np.array(graph_property)

    return graph_property

def rank_norm(dFC_mat, global_norm=True):
    '''
    dFC_mat = (n_time, n_region, n_region)
    dFC_mat_norm = rank_norm(dFC_mat)
    if global_norm=True, all time points ranked together, ow separately
    if dFC_mat = (n_region, n_region) -> dFC_mat_new = (n_region, n_region)
    '''
    dFC_mat_copy = deepcopy(dFC_mat)
    flag_dim = False
    if len(dFC_mat_copy.shape)<3:
        dFC_mat_copy = np.expand_dims(dFC_mat_copy, axis=0)
        flag_dim = True
    assert dFC_mat_copy.shape[1]==dFC_mat_copy.shape[2], \
        'dimension mismatch.'
    n_time = dFC_mat_copy.shape[0]
    n_region = dFC_mat_copy.shape[1]
    dFC_vecs = dFC_mat2vec(dFC_mat_copy) # (n_time, (n_region*(n_region-1))/2)
    if global_norm:
        dFC_vecs_flatten = dFC_vecs.flatten() # (n_time*(n_region*(n_region-1))/2,)
        dFC_vecs_flatten_ranked = stats.rankdata(dFC_vecs_flatten)
        dFC_vecs_ranked = dFC_vecs_flatten_ranked.reshape((n_time, -1)) # (n_time, (n_region*(n_region-1))/2)
        dFC_mat_ranked = dFC_vec2mat(dFC_vecs_ranked, N=n_region) # (n_time, n_region, n_region)
        dFC_mat_new = dFC_mat_ranked
    else:
        # normalize time point-wise
        dFC_vecs_new = list()
        for i, vec in enumerate(dFC_vecs):
            vec_ranked = stats.rankdata(vec) # (n_region*(n_region-1))/2,)
            dFC_vecs_new.append(vec_ranked)
        dFC_vecs_new = np.array(dFC_vecs_new) # (n_time, (n_region*(n_region-1))/2)
        dFC_mat_new = dFC_vec2mat(dFC_vecs_new, N=n_region) # (n_time, n_region, n_region)
    if flag_dim:
        dFC_mat_new = np.squeeze(dFC_mat_new) # (n_region, n_region)

    return dFC_mat_new


def rank_norm_dFC_dict(dFC_dict, global_norm=True):
    '''
    rank normalize dFC matrices in dFC_dict
    dFC_dict = {'TR1': FC_mat1, 'TR2': FC_mat2, ...}
    '''
    dFC_mat = np.array([dFC_dict[TR] for TR in dFC_dict])
    dFC_mat_new = rank_norm(dFC_mat, global_norm=global_norm)
    dFC_dict_new = {}
    for i, TR in enumerate(dFC_dict):
        dFC_dict_new[TR] = dFC_mat_new[i, :, :]
    return dFC_dict_new


def cat_data(X_t, N):
    '''
    X_t = (time, roi, roi)
    X_t is preferable to be ranked prior to cat_data
    '''
    X_t_new = list()
    for X in X_t:
        borders = np.linspace(1, np.max(X), N, endpoint=False)
        score_mat = np.zeros(X.shape)
        for border in borders:
            score_mat += X >= border
        X_t_new.append(score_mat)
    X_t_new = np.array(X_t_new)
    return X_t_new

def dFC_mask(dFC_mat, mask):
    '''
    dFC_mat and mask will be vectorized using dFC_mat2vec
    mask = (roi, roi)
    '''
    dFC_vecs = dFC_mat2vec(dFC_mat)
    mask_vec = dFC_mat2vec(mask)

    dFC_vec_new = list()
    for dFC_vec in dFC_vecs:
        dFC_vec_new.append(dFC_vec[mask_vec])
    dFC_vec_new = np.array(dFC_vec_new)

    return dFC_vec_new

#test
# toDo: use ssd.squareform
def dFC_mat2vec(C_t):
    '''
    C_t must be an array of matrices or a single matrix
    diagonal values will not be included. if you want to include 
    them set k=0
    if C_t is a single matrix, F will be one dim
    changing F will not change C_t
    '''
    if len(C_t.shape)==2:
        assert C_t.shape[0]==C_t.shape[1],\
            'C is not a square matrix'
        return C_t[np.triu_indices(C_t.shape[1], k=1)]

    F = list()
    for t in range(C_t.shape[0]):
        C = C_t[t, : , :]
        assert C.shape[0]==C.shape[1],\
            'C is not a square matrix'
        F.append(C[np.triu_indices(C_t.shape[1], k=1)])

    F = np.array(F)
    return F

#test
# toDo: use ssd.squareform
def dFC_vec2mat(F, N):
    '''
    diagonal values are set to 1.0
    F shape is (observations, features)
    '''
    C = list()
    iu = np.triu_indices(N, k=1)
    for i in range(F.shape[0]):
        K = np.zeros((N, N))
        K[iu] = F[i,:]
        K = K + K.T
        K = K + np.eye(N)
        C.append(K)
    C = np.array(C)
    return C

# test
def common_subj_lst(time_series_dict):
    SUBJECTs = None
    for session in time_series_dict:
        if SUBJECTs is None:
            SUBJECTs = time_series_dict[session].subj_id_lst
        else:
            SUBJECTs = intersection(SUBJECTs, time_series_dict[session].subj_id_lst)
    return SUBJECTs

def intersection(lst1, lst2): # input is a list 
    lst3 = [value for value in lst1 if value in lst2]
    return lst3

def TR_intersection(dFC_lst): # input is a list of dFC objs
    TRs_lst_old = dFC_lst[0].TR_array
    common_Fs = dFC_lst[0].TS_info['Fs']
    for dFC in dFC_lst:
        assert dFC.TS_info['Fs'] == common_Fs, \
            'Fs mismatch. Cannot find the common TRs'
            
        TRs_lst_new = intersection(TRs_lst_old, dFC.TR_array)
        TRs_lst_old = TRs_lst_new
    TRs_lst_old.sort()
    if len(TRs_lst_old)==0:
        print('No TR intersection.')
    return TRs_lst_old

def dFC_dict_slice(data, idx_lst):
    data_sliced = {}
    for i, k in enumerate(data):
        if i in idx_lst:
            data_sliced[k] = data[k]
    return data_sliced

def node_info2network(nodes_info):
    node_networks = []
    for info in nodes_info:
        if info[3]=='Network':
            continue
        node_networks.append(info[3])    
    return node_networks

def label2network(label):
    '''
    returns the network name of a label
    label format: Hemisphere_Network_ID
    '''
    return label[label.find('_')+1:label.find('_', label.find('_')+1)]

def node_labels2networks(node_labels):
    node_networks = []
    for label in node_labels:
        node_networks.append(label2network(label))
    return node_networks

def segment_FC(FC, node_networks):
    unique_node_networks = list(set(node_networks))
    segmented = np.zeros_like(FC)
    for network_i in unique_node_networks:
        node_id_i = [idx for idx, value in enumerate(node_networks) if value==network_i]
        for network_j in unique_node_networks:
            node_id_j = [idx for idx, value in enumerate(node_networks) if value==network_j]
            segmented[node_id_i[0]:node_id_i[-1]+1, node_id_j[0]:node_id_j[-1]+1] = np.mean(FC[node_id_i[0]:node_id_i[-1]+1, node_id_j[0]:node_id_j[-1]+1])
    return segmented
        
def segment_FC_dict(FC_dict, node_networks):
    segmented_dict = {}
    for key in FC_dict:
        segmented_dict[key] = segment_FC(FC_dict[key], node_networks)
    return segmented_dict
    
def visualize_conn_mat(C, axis=None, title='',
    cmap='seismic',
    V_MIN=None, V_MAX=None,
    node_networks=None,
    title_fontsize=18,
    loc_x=None, loc_y=None,
    ):
    '''
    C is (regions, regions)
    you can use loc_x and loc_y to set the location of the image
    loc_x and loc_y are lists of two elements, [start, end]
    '''

    if axis is None:
        fig, axis = plt.subplots(1, 1, figsize=(5, 5))

    if node_networks is None:
        axis.set_axis_off()

    if V_MAX is None:
        V_MAX = np.max(np.abs(C))
    if V_MIN is None:
        V_MIN = -1*V_MAX

    if loc_x is None or loc_y is None:
        im = axis.imshow(
            C, 
            interpolation='nearest', aspect='equal', cmap=cmap,    # 'viridis' or 'jet'
            vmin=V_MIN, vmax=V_MAX
        )
    else:
        im = axis.imshow(
            C, 
            interpolation='nearest', aspect='equal', cmap=cmap,    # 'viridis' or 'jet'
            vmin=V_MIN, vmax=V_MAX, 
            extent=[loc_x[0], loc_x[1], loc_y[0], loc_y[1]]
        )
    
    # cluster node networks
    if not node_networks is None:
        
        # finding unique network names wrt order
        network_names = []
        for node in node_networks:
            if not node in network_names:
                network_names.append(node)
        network_labels = [network_names.index(node) for node in node_networks]

        network_borders = np.argwhere(np.diff(network_labels)!=0)
        ticks_position = []
        last_line_position = 0
        for i in network_borders:
            # 0.5 is the visualization offset of imshow
            line_position = i[0]+1-0.5
            axis.axvline(x=line_position, color='k', linewidth=1)
            axis.axhline(y=line_position, color='k', linewidth=1)
            ticks_position.append((line_position+last_line_position)/2)
            last_line_position = line_position
        line_position = len(node_networks)+1-0.5
        ticks_position.append((line_position+last_line_position)/2)

        axis.set_xticks(ticks_position)
        axis.set_yticks(ticks_position)
        axis.set_xticklabels(network_names, rotation=90, fontsize=13)
        axis.set_yticklabels(network_names, fontsize=13)
    
    axis.set_title(title, fontdict={'fontsize': title_fontsize, 'fontweight': 'bold'})

    return im

def visualize_conn_mat_dict(data, title='', 
    cmap='seismic',
    normalize=False,
    disp_diag=True,
    label_dict={},
    save_image=False, output_root=None, axes=None, fig=None, 
    fix_lim=True, center_0=True, 
    node_networks=None, segmented=False 
    ):

    '''
    - data must be a dict of connectivity matrices
    sample:
    Suptitle1
        0.00 0.31 0.76 
        0.31 0.00 0.43 
        0.76 0.43 0.00 
    Suptitle1
        0.00 0.32 0.76 
        0.32 0.00 0.45 
        0.76 0.45 0.00 
    '''

    sns.set_context("paper", 
        font_scale=2.5, 
        rc={"lines.linewidth": 3.0}
    )

    sns.set_style('white')

    if node_networks is None:
        fig_width = 25*(len(data)/10)
    else:
        fig_width = 60*(len(data)/10)
    fig_height = 5

    fig_flag = True
    if axes is None or fig is None:
        fig_flag = False

    if not fig_flag:
        fig, axes = plt.subplots(1, len(data), figsize=(fig_width, fig_height), \
            facecolor='w', edgecolor='k')

    if not type(axes) is np.ndarray:
        axes = np.array([axes])

    if show_title:
        fig.suptitle(title, fontsize=20, y=0.98) #, fontsize=20, size=20

    axes = axes.ravel()

    # normalizing and scale
    conn_mats = list()
    V_MAX_all = None
    for i, key in enumerate(data):
        
        if segmented:
            C = data[key]
            if not disp_diag:
                C = np.multiply(C, 1-np.eye(len(C)))
                C = C + np.mean(C.flatten()) * np.eye(len(C))
            C = segment_FC(C, node_networks)
        else:
            C = data[key]

        if normalize:
            C = dFC_mat_normalize(C[None,:,:], global_normalization=False, threshold=0.0)[0]

        if (not disp_diag) and (not segmented):
            C = np.multiply(C, 1-np.eye(len(C)))
            C = C + np.mean(C.flatten()) * np.eye(len(C))

        if V_MAX_all is None:
            V_MAX_all = np.max(np.abs(C))
        else:
            V_MAX_all = max(V_MAX_all, np.max(np.abs(C)))

        conn_mats.append(C)
    conn_mats = np.array(conn_mats)

    if np.any(conn_mats<0) or center_0: 
        V_MIN = -1
        V_MAX = 1
    else: 
        V_MIN = 0
        V_MAX = 1

    if not fix_lim:
        V_MAX = V_MAX_all
        if np.any(conn_mats<0) or center_0:
            V_MIN = -1 * V_MAX_all
        else:
            V_MIN = 0

    # plot
    for i, key in enumerate(data):

        C = conn_mats[i,:,:]

        mat_title = key
        if key in label_dict:
            mat_title = label_dict[key]
        im = visualize_conn_mat(C, axis=axes[i], title=mat_title,
            cmap=cmap,
            V_MIN=V_MIN, V_MAX=V_MAX,
            node_networks=node_networks
            )
    if not fig_flag:
        fig.subplots_adjust(
            bottom=0.1, \
            top=0.85, \
            left=0.1, \
            right=0.9,
            # wspace=0.02, \
            # hspace=0.02\
        )

        if not node_networks is None:
            fig.subplots_adjust(
                wspace=0.55 
            )
            
    l, b, w, h = axes[-1].get_position().bounds
    if fig_flag:
        cb_ax = fig.add_axes([0.91, b, 0.007, h])
    else:
        cb_ax = fig.add_axes([0.91, b, 0.01, h])
    cbar = fig.colorbar(im, cax=cb_ax, shrink=0.8) # shrink=0.8??

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title.replace(" ", "_")+'.'+save_fig_format, 
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    # else:
    #     plt.show()


def visualize_conn_mat_2D_dict(data, title='', \
    cmap='seismic',\
    normalize=False,\
    disp_diag=True,\
    save_image=False, output_root=None, \
    fix_lim=True, center_0=True, \
    node_networks=None, segmented=False \
    ):

    '''
    - data must be a 2D dict of connectivity matrices
    sample:
    ROW1 (method_1)
        COLUMN1 (method_1)
            data[method_1][method_1]
                0.00 0.31 0.76 
                0.31 0.00 0.43 
                0.76 0.43 0.00 
        COLUMN2 (method_2)
            data[method_1][method_2]
                0.00 0.31 0.76 
                0.31 0.00 0.43 
                0.76 0.43 0.00 
    ROW2 (method_2)
        COLUMN1 (method_1)
            data[method_2][method_1]
                0.00 0.31 0.76 
                0.31 0.00 0.43 
                0.76 0.43 0.00 
    '''
    zip_measure_name = True
    sns.set_context("paper", 
        font_scale=3.5, 
        rc={"lines.linewidth": 3.0}
    )

    sns.set_style('white')

    if node_networks is None:
        fig_width = 30*(len(data)/10)
    else:
        fig_width = 55*(len(data)/10) + 4
    fig_height = fig_width * 1.0

    fig, axs = plt.subplots(len(data), len(data), figsize=(fig_width, fig_height), \
        facecolor='w', edgecolor='k')

    if not type(axs) is np.ndarray:
        axs = np.array([axs])

    if show_title:
        fig.suptitle(title, fontsize=25, y=0.98) #, fontsize=20, size=20

    # axs = axs.ravel()

    # normalizing and scale
    conn_mats = list()
    V_MAX_all = None
    for i, key_i in enumerate(data):
        for j, key_j in enumerate(data[key_i]):
            
            if segmented:
                C = segment_FC(data[key_i][key_j], node_networks)
            else:
                C = data[key_i][key_j]
            

            if normalize:
                C = dFC_mat_normalize(C[None,:,:], global_normalization=False, threshold=0.0)[0]

            if not disp_diag:
                C = np.multiply(C, 1-np.eye(len(C)))
                C = C + np.mean(C.flatten()) * np.eye(len(C))

            if V_MAX_all is None:
                V_MAX_all = np.max(np.abs(C))
            else:
                V_MAX_all = max(V_MAX_all, np.max(np.abs(C)))

            conn_mats.append(C)
    conn_mats = np.array(conn_mats)

    if np.any(conn_mats<0) or center_0: 
        V_MIN = -1
        V_MAX = 1
    else: 
        V_MIN = 0
        V_MAX = 1

    if not fix_lim:
        V_MAX = V_MAX_all
        if np.any(conn_mats<0) or center_0:
            V_MIN = -1 * V_MAX_all
        else:
            V_MIN = 0
    
    # plot
    axs_plotted = list()
    for i, key_i in enumerate(data):

        for j, key_j in enumerate(data[key_i]):

            if segmented:
                C = segment_FC(data[key_i][key_j], node_networks)
            else:
                C = data[key_i][key_j]

            if normalize:
                C = dFC_mat_normalize(C[None,:,:], global_normalization=False, threshold=0.0)[0]

            if not disp_diag:
                C = np.multiply(C, 1-np.eye(len(C)))
                C = C + np.mean(C.flatten()) * np.eye(len(C))

            if zip_measure_name:
                mat_title=zip_name(key_i)+'-'+zip_name(key_j)
            else:
                mat_title=key_i+' and '+key_j

            im = visualize_conn_mat(C, axis=axs[i][j], title=mat_title,
                cmap=cmap,
                V_MIN=V_MIN, V_MAX=V_MAX,
                node_networks=node_networks,
                title_fontsize=25
            )

            axs_plotted.append(axs[i][j])

    # remove extra subplots
    for ax in axs.ravel():
        if not ax in axs_plotted:
            ax.set_axis_off()
            ax.xaxis.set_tick_params(which='both', labelbottom=True)

    fig.subplots_adjust(
        bottom=0.1, \
        top=0.95, \
        left=0.1, \
        right=0.9,
        wspace=0.001, \
        hspace=0.4\
    )

    if not node_networks is None:
        fig.subplots_adjust(
            wspace=0.45, 
            hspace=0.50
        )
        
    l, b, w, h = axs[-1][-1].get_position().bounds
    if node_networks is None:
        cb_ax = fig.add_axes([0.91, 0.5-h/2, 0.007, h])
    else:
        cb_ax = fig.add_axes([0.91, 0.5-h/2, 0.015, h])
    cbar = fig.colorbar(im, cax=cb_ax, shrink=0.8) # shrink=0.8??

    if save_image:
        folder = output_root[:output_root.rfind('/')]
        if not os.path.exists(folder):
            os.makedirs(folder)
        plt.savefig(output_root+title.replace(" ", "_")+'.'+save_fig_format,
            dpi=fig_dpi, bbox_inches=fig_bbox_inches, pad_inches=fig_pad, format=save_fig_format
        ) 
        plt.close()
    else:
        plt.show()
        

def visualize_FCS(
        measure,
        normalize=True, fix_lim=True, 
        save_image=False, output_root=None
    ):
    
    if measure.FCS == []:
        return

    if normalize:
        D = dFC_dict_normalize(D=measure.FCS_dict, global_normalization=False)
    else:
        D = measure.FCS_dict

    fig_width = 45*(len(D)/10)
    fig_height = 8

    fig, axes = plt.subplots(2, len(D), figsize=(fig_width, fig_height), 
        facecolor='w', edgecolor='k')

    fig.subplots_adjust(
        bottom=0.1, 
        top=0.85, 
        left=0.1, 
        right=0.9,
        wspace=0.1, 
        hspace=0.6
    )

    # plot mean activity
    for i, mean_act in enumerate(measure.mean_act):
        # setting vmin=-vmax to make 0 correspond to white color
        max_activity = np.max(np.abs(mean_act))
        min_activity = -1*max_activity
        plot_markers(
            node_values=mean_act, 
            node_coords=measure.TS_info['nodes_locs'], 
            node_cmap='seismic', 
            node_vmin=min_activity, node_vmax=max_activity,
            display_mode='z', 
            colorbar=False, axes=axes[1, i]
        )

    # plot FC pattern
    node_networks = node_info2network(measure.TS_info['nodes_info'])

    visualize_conn_mat_dict(data=D, \
        node_networks=node_networks, \
        title=measure.measure_name+' FCS', \
        save_image=save_image, \
        axes=axes[0, :], fig=fig, 
        output_root=output_root, \
        disp_diag=False, \
        fix_lim=fix_lim \
    )

    fig.subplots_adjust(
        bottom=0.1, 
        top=0.85, 
        left=0.1, 
        right=0.9,
        wspace=0.1, 
        hspace=1.0
    )

'''
########## bundled brain graph visualizer ##########

cvsopts = dict(plot_height=400, plot_width=400)

def thresh_G(G, threshold):
    
    G_copy = deepcopy(G)
    
    if threshold > 1:
        labels = [d["weight"] for (u, v, d) in G_copy.edges(data=True)]
        labels.sort()
        threshold = labels[-1*threshold]
    
    ebunch = [(u, v) for u, v, d in G_copy.edges(data=True) if np.abs(d['weight']) < threshold]
    G_copy.remove_edges_from(ebunch)
        
    return G_copy

def nodesplot(nodes, name=None, canvas=None, cat=None):
    canvas = ds.Canvas(**cvsopts) if canvas is None else canvas
    # aggregator=None if cat is None else ds.count_cat(cat)
    # agg=canvas.points(nodes,'x','y',aggregator)
    aggc = canvas.points(nodes, 'x', 'y', ds.count_cat('cat'))   #ds.by('cat', ds.count())
        
    color_key = dict(cat_normal='#FF3333', cat_sig='#00FF00')
    
    return tf.spread(tf.shade(aggc, color_key=color_key), px=4, name=name)


def edgesplot(edges, name=None, canvas=None):
    canvas = ds.Canvas(**cvsopts) if canvas is None else canvas
    return tf.shade(canvas.line(edges, 'x','y', agg=ds.count()), name=name)
    
def graphplot(nodes, edges, name="", canvas=None, cat=None):
    
    if canvas is None:
        xr = nodes.x.min(), nodes.x.max()
        yr = nodes.y.min(), nodes.y.max()
        canvas = ds.Canvas(x_range=xr, y_range=yr, **cvsopts)
        
    np = nodesplot(nodes, name + " nodes", canvas, cat)
    ep = edgesplot(edges, name + " edges", canvas)
    return tf.stack(ep, np, how="over", name=name)

def ng(graph,name):
    graph.name = name
    return graph

def nx_layout(graph, view_degree=0, threshold=0):
    # layout = nx.circular_layout(graph)
    
    # Get node positions
    pos = nx.get_node_attributes(graph, 'pos')
    for key in pos:
        if view_degree==0:    
            pos[key] = pos[key][:2]
        if view_degree==1:    
            pos[key] = pos[key][1:3]
        if view_degree==2:    
            pos[key] = pos[key][[0, 2]]
            
    # layout = pos
            
    cat = list()
    for key in graph.nodes():
        cat.append('cat_normal')
        # if key in find_sig_nodes(graph):   #
        #     cat.append( 'cat_sig')
        # else:
        #     cat.append('cat_normal')
            
    data = [[node]+pos[node].tolist()+[cat[i]] for i, node in enumerate(graph.nodes)]

    nodes = pd.DataFrame(data, columns=['id', 'x', 'y','cat'])
    nodes.set_index('id', inplace=True)
    nodes["cat"]=nodes["cat"].astype("category")

    graph_copy = thresh_G(graph, threshold=threshold)

    edges = pd.DataFrame(list(graph_copy.edges), columns=['source', 'target'])
    return nodes, edges

def nx_plot(graph, name="", view_degree=0, threshold=0):
    # print(graph.name, len(graph.edges))
    nodes, edges = nx_layout(graph, view_degree=view_degree, threshold=threshold)
    
    direct = connect_edges(nodes, edges)
    bundled_bw005 = hammer_bundle(nodes, edges)
    bundled_bw030 = hammer_bundle(nodes, edges, initial_bandwidth=0.30)
    bundled_bw100 = hammer_bundle(nodes, edges, initial_bandwidth=1)

    return [graphplot(nodes, direct,         graph.name, cat=None),
            graphplot(nodes, bundled_bw005, "Bundled bw=0.05", cat=None),
            graphplot(nodes, bundled_bw030, "Bundled bw=0.30", cat=None),
            graphplot(nodes, bundled_bw100, "Bundled bw=1.00", cat=None)]

def batch_Adj2Net(FCS, nodes_info, is_digraph=False):
  
    np.fill_diagonal(FCS, 0)
    if is_digraph:
        G = nx.from_numpy_array(FCS, create_using=nx.DiGraph)
    else:
        G = nx.from_numpy_array(FCS)

    mapping = {}
    for i, node_info in enumerate(nodes_info):
        mapping[i] = node_info[4]
    G = nx.relabel_nodes(G, mapping)

    return G

def set_locs_G(G, locs):
    
    G_copy = deepcopy(G)
    
    pos = nx.circular_layout(G_copy) 

    for i, key in enumerate(pos):
        pos[key] = locs[i]
        
    nx.set_node_attributes(G_copy, pos, "pos") 
      
    
    return G_copy 

def visulize_brain_graph(FCS, nodes_info, locs, num_edges2show, \
    title='', save_image=True, output_root=None \
    ):
    
    # EXAMPLE:
    # visulize_brain_graph(measure.FCS_dict[FCS], measure.TS_info['nodes_info'], \
    # measure.TS_info['nodes_locs'], num_edges2show=100, \
    # title=FCS+'_'+measure.measure_name, save_image=save_image, output_root=output_root \
    # )
    
    G = batch_Adj2Net(FCS=FCS, nodes_info=nodes_info, is_digraph=False)
    G = set_locs_G(G, locs=locs)   
    plots = [nx_plot(ng(G, name="dFC"), view_degree=0, threshold=num_edges2show)]

    if save_image:
        ds.utils.export_image(img=plots[0][2], filename=title+'_bundle_', 
                        fmt=".png", background='black', 
                        export_path=output_root)
    
    # return plots[0][0]


##############################
'''

def dFC_dict_normalize(D, global_normalization=False, threshold=0.0):

    C = list()
    for key in D:
        C.append(D[key])
    C = np.array(C)

    C_z = dFC_mat_normalize(C, \
        global_normalization=global_normalization, \
        threshold=threshold \
    )

    D_z = {}
    for i, key in enumerate(D):
        D_z[key] = C_z[i,:,:]

    return D_z

def dFC_mat_normalize(C_t, global_normalization=False, threshold=0.0):

    # threshold is ratio of connections wanted to be zero
    C_t_z = deepcopy(C_t)
    if len(C_t_z.shape)<3:
        C_t_z = np.expand_dims(C_t_z, axis=0)

    if global_normalization:

        # transform the whole abs(dFC mat) to [0, 1] 

        signs = np.sign(C_t_z)
        C_t_z = np.abs(C_t_z)

        miN = list()
        for i in range(C_t_z.shape[0]):
            slice = C_t_z[i,:,:]
            slice_non_diag = slice[np.where(~np.eye(slice.shape[0],dtype=bool))]
            miN.append(np.min(slice_non_diag))

        C_t_z = C_t_z - np.min(miN)

        maX = list()
        for i in range(C_t_z.shape[0]):
            slice = C_t_z[i,:,:]
            slice_non_diag = slice[np.where(~np.eye(slice.shape[0],dtype=bool))]
            maX.append(np.max(slice_non_diag))

        if np.max(maX) != 0:
            C_t_z = np.divide(C_t_z, np.max(maX))

        # thresholding
        d = deepcopy(np.ravel(C_t_z))
        d.sort()
        new_threshold = d[int(threshold*len(d))]
        C_t_z = np.multiply(C_t_z, (C_t_z>=new_threshold))
        C_t_z = np.multiply(C_t_z, signs)

    else:

        # transform abs of each time slice to [0, 1]

        signs = np.sign(C_t_z)
        C_t_z = np.abs(C_t_z)
        
        for i in range(C_t_z.shape[0]):
            slice = C_t_z[i,:,:]
            slice_non_diag = slice[np.where(~np.eye(slice.shape[0],dtype=bool))]
            slice = slice - np.min(slice_non_diag)
            slice_non_diag = slice[np.where(~np.eye(slice.shape[0],dtype=bool))]
            if np.max(slice_non_diag) != 0:
                slice = np.divide(slice, np.max(slice_non_diag))

            # thresholding
            d = deepcopy(np.ravel(slice))
            d.sort()
            new_threshold = d[int(threshold*len(d))]
            slice = np.multiply(slice, (slice>=new_threshold))

            C_t_z[i,:,:] = slice

        C_t_z = np.multiply(C_t_z, signs)

    # removing self connections
    for i in range(C_t_z.shape[1]):
        C_t_z[:, i, i] = np.mean(C_t_z) # ?????????????????

    return C_t_z

def print_mat(mat, s=0):
    if len(mat.shape)==1:
        mat = np.expand_dims(mat, axis=0)
    for i in mat:
        print('\t'*s,  end=" ")
        for j in i:
            print("{:.2f}".format(j), end=" ")
        print()

def print_dict(t, s=0):
    if not isinstance(t,dict) and not isinstance(t,list):
        if isinstance(t,np.ndarray):
            print_mat(t, s)
        else:
            if isinstance(t,float):
                print('\t'*s+"{:.2f}".format(t))
            else:
                print('\t'*s+str(t))
    else:
        for key in t:
            print('\t'*s+str(key))
            if not isinstance(t,list):
                print_dict(t[key],s+1)
